"""
Defines import/export functionality
"""

from __future__ import annotations
import xml.etree.ElementTree as ET
from os import PathLike, listdir
from pathlib import Path
from random import shuffle
from typing import Any, Callable, Iterable, Sequence

from attrs import define, field

from negmas.helpers.inout import dump, load
from negmas.helpers.types import get_full_type_name
from negmas.outcomes.outcome_space import make_os
from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction
from negmas.preferences.ops import ScenarioStats, calc_scenario_stats
from negmas.sao.mechanism import SAOMechanism
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

from .mechanisms import Mechanism
from .negotiators import Negotiator
from .outcomes import (
    CartesianOutcomeSpace,
    Issue,
    issues_from_genius,
    issues_from_geniusweb,
)
from .preferences import (
    DiscountedUtilityFunction,
    UtilityFunction,
    conflict_level,
    make_discounted_ufun,
    nash_points,
    opposition_level,
    pareto_frontier,
    winwin_level,
)
from .preferences.value_fun import TableFun

__all__ = [
    "Scenario",
    "load_genius_domain",
    "load_genius_domain_from_folder",
    "find_genius_domain_and_utility_files",
    "load_geniusweb_domain",
    "load_geniusweb_domain_from_folder",
    "find_geniusweb_domain_and_utility_files",
    "get_domain_issues",
]

STATS_MAX_CARDINALITY = 10_000_000_000
GENIUSWEB_UFUN_TYPES = ("LinearAdditiveUtilitySpace",)


@define
class Scenario:
    """
    A class representing a negotiation domain
    """

    outcome_space: CartesianOutcomeSpace
    ufuns: tuple[UtilityFunction, ...]
    mechanism_type: type[Mechanism] | None = SAOMechanism
    mechanism_params: dict = field(factory=dict)

    @property
    def issues(self) -> tuple[Issue, ...]:
        return self.outcome_space.issues

    def plot(self, **kwargs):
        from negmas.plots.util import plot_2dutils

        return plot_2dutils(
            [],
            self.ufuns,  #
            tuple(_.name for _ in self.ufuns),  #
            offering_negotiators=tuple(_.name for _ in self.ufuns),  #
            issues=self.outcome_space.issues,  #
            **kwargs,
        )

    def to_genius_files(self, domain_path: Path, ufun_paths: list[Path]):
        """
        Save domain and ufun files to the `path` as XML.
        """
        domain_path = Path(domain_path)
        ufun_paths = [Path(_) for _ in ufun_paths]
        if len(self.ufuns) != len(ufun_paths):
            raise ValueError(
                f"I have {len(self.ufuns)} ufuns but {len(ufun_paths)} paths were passed!!"
            )
        domain_path.parent.mkdir(parents=True, exist_ok=True)
        self.outcome_space.to_genius(domain_path)
        for ufun, path in zip(self.ufuns, ufun_paths):
            ufun.to_genius(path, issues=self.issues)
        return self

    def to_genius_folder(self, path: Path):
        """
        Save domain and ufun files to the `path` as XML.
        """
        path.mkdir(parents=True, exist_ok=True)
        domain_name = (
            self.outcome_space.name.split("/")[-1]
            if self.outcome_space.name
            else "domain"
        )
        ufun_names = [_.name.split("/")[-1] for _ in self.ufuns]
        self.outcome_space.to_genius(path / domain_name)
        for ufun, name in zip(self.ufuns, ufun_names):
            ufun.to_genius(path / name, issues=self.issues)
        return self

    @property
    def n_negotiators(self) -> int:
        return len(self.ufuns)

    @property
    def n_issues(self) -> int:
        return len(self.outcome_space.issues)

    @property
    def issue_names(self) -> list[str]:
        return self.outcome_space.issue_names

    def to_numeric(self) -> Scenario:
        """
        Forces all issues in the domain to become numeric

        Remarks:
            - maps the agenda and ufuns to work correctly together
        """
        raise NotImplementedError()

    def _randomize(self) -> Scenario:
        """
        Randomizes the outcomes in a single issue scneario
        """
        shuffle(self.outcome_space.issues[0].values)
        return self

    def to_single_issue(
        self, numeric=False, stringify=True, randomize=False
    ) -> Scenario:
        """
        Forces the domain to have a single issue with all possible outcomes

        Args:
            numeric: If given, the output issue will be a `ContiguousIssue` otherwise it will be a `DiscreteCategoricalIssue`
            stringify:  If given, the output issue will have string values. Checked only if `numeric` is `False`
            randomize: Randomize outcome order when creating the single issue

        Remarks:
            - maps the agenda and ufuns to work correctly together
            - Only works if the outcome space is finite
        """
        if (
            hasattr(self.outcome_space, "issues")
            and len(self.outcome_space.issues) == 1
        ):
            return self if not randomize else self._randomize()
        outcomes = list(self.outcome_space.enumerate_or_sample())
        sos = self.outcome_space.to_single_issue(numeric, stringify)
        ufuns = []
        souts = list(sos.issues[0].all)
        for u in self.ufuns:
            if isinstance(u, DiscountedUtilityFunction):
                usave = u
                v = u.ufun
                while isinstance(v, DiscountedUtilityFunction):
                    u, v = v, v.ufun
                u.ufun = LinearAdditiveUtilityFunction(
                    values=(
                        TableFun(dict(zip(souts, [v(_) for _ in outcomes]))),
                    ),  #  (The error comes from having LRU cach for table ufun's minmax which should be OK)
                    bias=0.0,
                    reserved_value=v.reserved_value,
                    name=v.name,
                    outcome_space=sos,
                )
                ufuns.append(usave)
                continue
            ufuns.append(
                LinearAdditiveUtilityFunction(
                    values=(
                        TableFun(dict(zip(souts, [u(_) for _ in outcomes]))),
                    ),  #  (The error comes from having LRU cach for table ufun's minmax which should be OK)
                    bias=0.0,
                    reserved_value=u.reserved_value,
                    name=u.name,
                    outcome_space=sos,
                )
            )
        self.ufuns = tuple(ufuns)
        self.outcome_space = sos
        return self if not randomize else self._randomize()

    def make_session(
        self,
        negotiators: Callable[[], Negotiator]
        | type[Negotiator]
        | list[Negotiator]
        | tuple[Negotiator, ...]
        | None = None,
        n_steps: int | float | None = None,
        time_limit: float | None = None,
        roles: list[str] | None = None,
        raise_on_failure_to_enter: bool = True,
        **kwargs,
    ):
        """
        Generates a ready to run mechanism session for this domain.
        """
        if not self.mechanism_type:
            raise ValueError(
                "Cannot create the domain because it has no `mechanism_type`"
            )

        args = self.mechanism_params
        args.update(kwargs)
        m = self.mechanism_type(
            outcome_space=self.outcome_space,
            n_steps=n_steps,
            time_limit=time_limit,
            **args,
        )
        if not negotiators:
            return m
        negs: list[Negotiator]
        if not isinstance(negotiators, Iterable):
            negs = [
                negotiators(
                    name=ufun.name.split("/")[-1]  # type: ignore We trust that the class given is a negotiator and has a name
                    .replace(".xml", "")
                    .replace(".yml", "")
                )
                for ufun in self.ufuns
            ]
        else:
            negs = list(negotiators)
        if len(self.ufuns) != len(negs) or len(negs) < 1:
            raise ValueError(
                f"Invalid ufuns ({self.ufuns}) or negotiators ({negotiators})"
            )
        if not roles:
            roles = ["negotiator"] * len(negs)
        for n, r, u in zip(negs, roles, self.ufuns):
            added = m.add(n, preferences=u, role=r)
            if not added and raise_on_failure_to_enter:
                raise ValueError(
                    f"{n.name} (of type {get_full_type_name(n.__class__)}) failed to enter the negotiation {m.name}"
                )

        return m

    def scale_min(self, to: float = 1.0) -> Scenario:
        """Normalizes a utility function to the given range

        Args:
            ufun: The utility function to normalize
            outcomes: A collection of outcomes to normalize for
            rng: range to normalize to. Default is [0, 1]
            levels: Number of levels to use for discretizing continuous issues (if any)
            max_cardinality: Maximum allowed number of outcomes resulting after all discretization is done
        """
        self.ufuns = tuple(_.scale_min(to) for _ in self.ufuns)  #  The type is correct
        return self

    def scale_max(self, to: float = 1.0) -> Scenario:
        """Normalizes a utility function to the given range

        Args:
            ufun: The utility function to normalize
            outcomes: A collection of outcomes to normalize for
            rng: range to normalize to. Default is [0, 1]
            levels: Number of levels to use for discretizing continuous issues (if any)
            max_cardinality: Maximum allowed number of outcomes resulting after all discretization is done
        """
        self.ufuns = tuple(_.scale_max(to) for _ in self.ufuns)  #  The type is correct
        return self

    def normalize(self, to: tuple[float, float] = (0.0, 1.0)) -> Scenario:
        """Normalizes a utility function to the given range

        Args:
            rng: range to normalize to. Default is [0, 1]
        """
        self.ufuns = tuple(_.normalize(to) for _ in self.ufuns)  #  The type is correct
        return self

    def is_normalized(
        self,
        to: tuple[float | None, float | None] = (None, 1.0),
        positive: bool = True,
        eps: float = 1e-6,
    ) -> bool:
        """Checks that all ufuns are normalized in the given range"""
        mnmx = [_.minmax() for _ in self.ufuns]
        return all(
            (to[0] is None or abs(a - to[0]) < eps)
            and (to[1] is None or abs(b - to[1]) < eps)
            and (not positive or a >= -eps)
            for a, b in mnmx
        )

    def discretize(self, levels: int = 10):
        """Discretize all issues"""
        self.outcome_space = self.outcome_space.to_discrete(levels)
        return self

    def remove_discounting(self):
        """Removes discounting from all ufuns"""
        ufuns = []
        for u in self.ufuns:
            while isinstance(u, DiscountedUtilityFunction):
                u = u.ufun
            ufuns.append(u)
        self.ufuns = tuple(ufuns)
        return self

    def remove_reserved_values(self, r: float = float("-inf")):
        """Removes reserved values from all ufuns replaacing it with `r`"""
        for u in self.ufuns:
            u.reserved_value = r
        return self

    def calc_stats(self) -> ScenarioStats:
        return calc_scenario_stats(self.ufuns)

    def calc_extra_stats(
        self, max_cardinality: int = STATS_MAX_CARDINALITY
    ) -> dict[str, Any]:
        """
        Calculates and returns several stats corresponding to the domain

        Args:
            max_cardinality (int): The maximum number of outcomes considered when calculating the stats.

        Returns:
            A dictionary with the compiled stats
        """
        outcome_space, ufuns = self.outcome_space, self.ufuns
        outcomes = tuple(
            outcome_space.enumerate_or_sample(max_cardinality=max_cardinality)
        )
        frontier_utils, frontier_indices = pareto_frontier(
            ufuns, outcomes=outcomes, sort_by_welfare=True
        )
        frontier_outcomes = tuple(
            outcomes[_] for _ in frontier_indices if _ is not None
        )
        pts = nash_points(ufuns, frontier_utils, outcome_space=outcome_space)
        nash_utils, nash_indx = pts[0] if pts else (None, None)
        nash_outcome = frontier_outcomes[nash_indx] if nash_indx else None
        minmax = [u.minmax() for u in ufuns]
        nu, no, ol, cl, wl, fu, fo = (
            dict(),
            dict(),
            dict(),
            dict(),
            dict(),
            dict(),
            dict(),
        )
        opposition = opposition_level(
            ufuns,
            max_utils=tuple(_[1] for _ in minmax),  #
            outcomes=outcomes,
            max_tests=max_cardinality,
        )
        for i, u1 in enumerate(ufuns):
            if not u1:
                continue
            for j, u2 in enumerate(ufuns[i + 1 :]):
                if not u2:
                    continue
                us = (u1, u2)
                fu_, findx = pareto_frontier(
                    us, outcomes=outcomes, sort_by_welfare=True
                )
                foutcomes_ = [outcomes[i] for i in findx]
                fu[(u1.name, u2.name)], fo[(u1.name, u2.name)] = (
                    fu_,
                    [outcomes[_] for _ in findx],
                )
                pts = nash_points((u1, u2), fu_, outcomes=outcomes)
                nu[(u1.name, u2.name)], nindx = pts[0] if pts else (None, None)
                no[(u1.name, u2.name)] = foutcomes_[nindx] if nindx else None
                ol[(u1.name, u2.name)] = opposition_level(
                    (u1, u2),
                    outcomes=outcomes,
                    max_utils=(minmax[i][1], minmax[j][1]),
                    max_tests=max_cardinality,
                )
                cl[(u1.name, u2.name)] = conflict_level(
                    u1, u2, outcomes=outcomes, max_tests=max_cardinality
                )
                wl[(u1.name, u2.name)] = winwin_level(
                    u1, u2, outcomes=outcomes, max_tests=max_cardinality
                )

        return dict(
            frontier_utils=frontier_utils,
            frontier_outcomes=frontier_outcomes,
            nash_utils=nash_utils,
            nash_outcome=nash_outcome,
            opposition_level=opposition,
            bilateral_nash_utils=nu,
            bilateral_nash_outcome=no,
            bilateral_conflict_level=cl,
            bilateral_opposition_level=ol,
            bilateral_winwin_levl=wl,
            bilateral_frontier_utils=fu,
            bilateral_frontier_outcomes=fo,
        )

    def serialize(self) -> dict[str, Any]:
        """
        Converts the current scenario into a serializable dict.

        Remarks:
            Rturns a dictionary with the following keys:
                - domain: The agenda/outcome-space
                - ufuns: A list of utility functions
        """

        def get_name(x, default):
            if not x:
                return str(default)
            return x.split("/")[-1].replace(".xml", "")

        def adjust(
            d,
            default_name,
            remove_dunder=False,
            adjust_name=True,
            ignored=("id", "n_values", "outcome_space"),
            rename={PYTHON_CLASS_IDENTIFIER: "type"},
        ):
            if isinstance(d, list) or isinstance(d, tuple):
                return [
                    adjust(_, default_name, remove_dunder, adjust_name, ignored)
                    for _ in d
                ]
            if not isinstance(d, dict):
                return d
            if adjust_name and "name" in d:
                d["name"] = get_name(d["name"], default_name)
            if d.get(PYTHON_CLASS_IDENTIFIER, "").startswith("negmas."):
                d[PYTHON_CLASS_IDENTIFIER] = d[PYTHON_CLASS_IDENTIFIER].split(".")[-1]
            for old, new in rename.items():
                if old in d.keys():
                    d[new] = d[old]
                    del d[old]
            for i in ignored:
                if i in d.keys():
                    del d[i]
            for k, v in d.items():
                d[k] = adjust(
                    v,
                    default_name,
                    remove_dunder=remove_dunder,
                    adjust_name=False,
                    ignored=ignored,
                )
            if not remove_dunder:
                return d
            d = {k: v for k, v in d.items() if not k.startswith("__")}
            return d

        domain = adjust(
            serialize(self.outcome_space, shorten_type_field=True, add_type_field=True),
            "domain",
        )
        ufuns = [
            adjust(serialize(u, shorten_type_field=True, add_type_field=True), i)
            for i, u in enumerate(self.ufuns)
        ]
        return dict(domain=domain, ufuns=ufuns)

    def to_yaml(self, folder: Path | str) -> None:
        """
        Saves the scenario as yaml
        Args:
            folder: The destination path
        """
        self.dumpas(folder, "yml")

    def to_json(self, folder: Path | str) -> None:
        """
        Saves the scenario as json
        Args:
            folder: The destination path
        """
        self.dumpas(folder, "json")

    def dumpas(self, folder: Path | str, type="yml", compact: bool = False) -> None:
        """
        Dumps the scenario in the given file format.
        """
        folder = Path(folder)
        folder.mkdir(parents=True, exist_ok=True)
        serialized = self.serialize()
        dump(serialized["domain"], folder / f"{serialized['domain']['name']}.{type}")
        for u in serialized["ufuns"]:
            dump(u, folder / f"{u['name']}.{type}", sort_keys=True, compact=compact)

    @staticmethod
    def from_genius_folder(
        path: PathLike | str,
        ignore_discount=False,
        ignore_reserved=False,
        safe_parsing=True,
    ) -> Scenario | None:
        return load_genius_domain_from_folder(
            folder_name=str(path),
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            safe_parsing=safe_parsing,
        )

    @classmethod
    def load(cls, folder: Path | str, safe_parsing=False) -> Scenario | None:
        """
        Loads the scenario from a folder with supported formats: XML, YML
        """
        for finder, loader in (
            (find_domain_and_utility_files_yaml, cls.from_yaml_folder),
            (find_domain_and_utility_files_xml, cls.from_genius_folder),
            (find_domain_and_utility_files_geniusweb, cls.from_geniusweb_folder),
        ):
            domain, _ = finder(folder)
            if domain is not None:
                return loader(folder, safe_parsing=safe_parsing)

    @classmethod
    def is_loadable(cls, path: PathLike | str):
        if not Path(path).is_dir():
            return False
        for finder in (
            find_domain_and_utility_files_yaml,
            find_domain_and_utility_files_xml,
            find_domain_and_utility_files_geniusweb,
        ):
            d, _ = finder(Path(path))
            if d is not None:
                return True
        return False

    @staticmethod
    def from_genius_files(
        domain: PathLike,
        ufuns: Iterable[PathLike],
        ignore_discount=False,
        ignore_reserved=False,
        safe_parsing=True,
    ) -> Scenario | None:
        return load_genius_domain(
            domain,
            [_ for _ in ufuns],
            safe_parsing=safe_parsing,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            normalize_utilities=False,
            normalize_max_only=False,
        )

    @staticmethod
    def from_geniusweb_folder(
        path: PathLike | str,
        ignore_discount=False,
        ignore_reserved=False,
        use_reserved_outcome=False,
        safe_parsing=True,
    ) -> Scenario | None:
        return load_geniusweb_domain_from_folder(
            folder_name=str(path),
            ignore_discount=ignore_discount,
            use_reserved_outcome=use_reserved_outcome,
            ignore_reserved=ignore_reserved,
            safe_parsing=safe_parsing,
        )

    @staticmethod
    def from_geniusweb_files(
        domain: PathLike,
        ufuns: Iterable[PathLike],
        ignore_discount=False,
        ignore_reserved=False,
        use_reserved_outcome=False,
        safe_parsing=True,
    ) -> Scenario | None:
        return load_geniusweb_domain(
            domain,
            [_ for _ in ufuns],
            safe_parsing=safe_parsing,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            use_reserved_outcome=use_reserved_outcome,
            normalize_utilities=False,
            normalize_max_only=False,
        )

    @classmethod
    def from_yaml_folder(
        cls,
        path: PathLike | str,
        ignore_discount=False,
        ignore_reserved=False,
        safe_parsing=True,
    ) -> Scenario | None:
        domain, ufuns = find_domain_and_utility_files_yaml(path)
        if not domain:
            return None
        return cls.from_yaml_files(
            domain=domain,
            ufuns=ufuns,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            safe_parsing=safe_parsing,
        )

    @classmethod
    def from_yaml_files(
        cls,
        domain: PathLike,
        ufuns: Iterable[PathLike],
        ignore_discount=False,
        ignore_reserved=False,
        safe_parsing=True,
    ) -> Scenario | None:
        _ = safe_parsing  # yaml parsing is always safe

        def adjust_type(d: dict, base: str = "negmas", domain=None) -> dict:
            d["type"] = f"{base}.{d['type']}"
            if domain is not None:
                d["outcome_space"] = domain
            return d

        os = deserialize(
            adjust_type(load(domain)),
            python_class_identifier="type",
            base_module="negmas",
        )
        utils = tuple(
            deserialize(
                adjust_type(load(fname), domain=os),
                python_class_identifier="type",
                base_module="negmas",
            )
            for fname in ufuns
        )

        # d = load(domain)
        # type_ = d.pop("type", "")
        # assert (
        #     "OutcomeSpace" in type_
        # ), f"Unknown type or no type for domain file: {domain=}\n{d=}"
        # type_ = f"negmas.outcomes.{type_}"
        # d["issues"]
        # os = instantiate(type_, **d)
        # utils = []
        # for fname in ufuns:
        #     d = load(fname)
        #     type_ = d.pop("type", "")
        #     assert (
        #         "Fun" in type_
        #     ), f"Unknown type or no type for ufun file: {domain=}\n{d=}"
        #     type_ = f"negmas.preferences.{type_}"
        #     utils.append(instantiate(type_, **d))
        assert isinstance(os, CartesianOutcomeSpace)
        s = Scenario(outcome_space=os, ufuns=tuple(utils))  # type: ignore
        if s and ignore_discount:
            s = s.remove_discounting()
        if s and ignore_reserved:
            s = s.remove_reserved_values()
        return s


def get_domain_issues(
    domain_file_name: PathLike | str,
    n_discretization: int | None = None,
    safe_parsing=False,
) -> Sequence[Issue] | None:
    """
    Returns the issues of a given XML domain (Genius Format)

    Args:
        domain_file_name: Name of the file
        n_discretization: Number of discrete levels per continuous variable.
        max_cardinality: Maximum number of outcomes in the outcome space after discretization. Used only if `n_discretization` is given.
        safe_parsing: Apply more checks while parsing

    Returns:
        List of issues

    """
    issues = None
    if domain_file_name is not None:
        issues, _ = issues_from_genius(
            domain_file_name,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )
    return issues


def load_genius_domain(
    domain_file_name: PathLike,
    utility_file_names: Iterable[PathLike] | None = None,
    ignore_discount=False,
    ignore_reserved=False,
    safe_parsing=True,
    **kwargs,
) -> Scenario:
    """
    Loads a genius domain, creates appropriate negotiators if necessary

    Args:
        domain_file_name: XML file containing Genius-formatted domain spec
        utility_file_names: XML files containing Genius-fromatted ufun spec
        ignore_reserved: Sets the reserved_value of all ufuns to -inf
        ignore_discount: Ignores discounting
        safe_parsing: Applies more stringent checks during parsing

    Returns:
        A `Domain` ready to run

    """

    issues = None
    if domain_file_name is not None:
        issues, _ = issues_from_genius(domain_file_name, safe_parsing=safe_parsing)

    agent_info = []
    if utility_file_names is None:
        utility_file_names = []
    for ufname in utility_file_names:
        try:
            utility, discount_factor = UtilityFunction.from_genius(
                file_name=ufname,
                issues=issues,
                safe_parsing=safe_parsing,
                ignore_discount=ignore_discount,
                ignore_reserved=ignore_reserved,
                name=str(ufname),
            )
        except Exception as e:
            raise OSError(
                f"Ufun named {Path(ufname).name} cannot be read: {e.__class__.__name__}({e})"
            )

        agent_info.append(
            {
                "ufun": utility,
                "ufun_name": ufname,
                "reserved_value_func": utility.reserved_value
                if utility is not None
                else float("-inf"),
                "discount_factor": discount_factor,
            }
        )
    if domain_file_name is not None:
        kwargs["dynamic_entry"] = False
        kwargs["max_n_agents"] = None
        if not ignore_discount:
            for info in agent_info:
                info["ufun"] = (
                    info["ufun"]
                    if info["discount_factor"] is None or info["discount_factor"] == 1.0
                    else make_discounted_ufun(
                        ufun=info["ufun"],
                        discount_per_round=info["discount_factor"],
                        power_per_round=1.0,
                    )
                )
    if issues is None:
        raise ValueError(f"Could not load domain {domain_file_name}")

    return Scenario(
        outcome_space=make_os(issues, name=str(domain_file_name)),  #
        ufuns=tuple(_["ufun"] for _ in agent_info),  #
    )


def load_genius_domain_from_folder(
    folder_name: str | PathLike,
    ignore_reserved=False,
    ignore_discount=False,
    safe_parsing=False,
    **kwargs,
) -> Scenario:
    """
    Loads a genius domain from a folder. See ``load_genius_domain`` for more details.

    Args:
        folder_name: A folder containing one XML domain file and one or more ufun files in Genius format
        ignore_reserved: Sets the reserved_value of all ufuns to -inf
        ignore_discount: Ignores discounting
        safe_parsing: Applies more stringent checks during parsing
        kwargs: Extra arguments to pass verbatim to SAOMechanism constructor

    Returns:
        A domain ready for `make_session`

    Examples:

        >>> import pkg_resources
        >>> from negmas import load_genius_domain_from_folder

        Try loading and running a domain with predetermined agents:
        >>> domain = load_genius_domain_from_folder(
        ...     pkg_resources.resource_filename("negmas", resource_name="tests/data/Laptop")
        ... )


        Try loading a domain and check the resulting ufuns
        >>> domain = load_genius_domain_from_folder(
        ...     pkg_resources.resource_filename("negmas", resource_name="tests/data/Laptop")
        ... )

        >>> domain.n_issues, domain.n_negotiators
        (3, 2)

        >>> [type(_) for _ in domain.ufuns]
        [<class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>, <class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>]

        Try loading a domain forcing a single issue space
        >>> domain = load_genius_domain_from_folder(
        ...     pkg_resources.resource_filename("negmas", resource_name="tests/data/Laptop")
        ... ).to_single_issue()
        >>> domain.n_issues, domain.n_negotiators
        (1, 2)
        >>> [type(_) for _ in domain.ufuns]
        [<class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>, <class 'negmas.preferences.crisp.linear.LinearAdditiveUtilityFunction'>]


        Try loading a domain with nonlinear ufuns:
        >>> folder_name = pkg_resources.resource_filename(
        ...     "negmas", resource_name="tests/data/10issues"
        ... )
        >>> domain = load_genius_domain_from_folder(folder_name)
        >>> print(domain.n_issues)
        10
        >>> print(domain.n_negotiators)
        2
        >>> print([type(u) for u in domain.ufuns])
        [<class 'negmas.preferences.crisp.nonlinear.HyperRectangleUtilityFunction'>, <class 'negmas.preferences.crisp.nonlinear.HyperRectangleUtilityFunction'>]
        >>> u = domain.ufuns[0]
        >>> print(u.outcome_ranges[0])
        {1: (7.0, 9.0), 3: (2.0, 7.0), 5: (0.0, 8.0), 8: (0.0, 7.0)}

        >>> print(u.mappings[0])
        97.0
        >>> print(u([0.0] * domain.n_issues))
        0
        >>> print(u([0.5] * domain.n_issues))
        186.0
    """
    folder_name = str(folder_name)
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = Path(full_name)
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    if domain_file_name is None:
        raise ValueError("Cannot find a domain file")
    return load_genius_domain(
        domain_file_name=domain_file_name,
        utility_file_names=utility_file_names,
        safe_parsing=safe_parsing,
        ignore_reserved=ignore_reserved,
        ignore_discount=ignore_discount,
        **kwargs,
    )


def find_domain_and_utility_files_yaml(
    folder_name,
) -> tuple[PathLike | None, list[PathLike]]:
    """Finds the domain and utility_function files in a folder"""
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    folder_name = str(folder_name)
    for f in files:
        if not f.endswith(".yml") and not f.endswith(".yaml"):
            continue
        full_name = folder_name + "/" + f
        data = load(full_name)
        if data and "OutcomeSpace" in data.get("type", ""):
            domain_file_name = full_name
        elif data and ("fun" in data.get("type", "").lower()):
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names


def find_domain_and_utility_files_geniusweb(
    folder_name,
) -> tuple[PathLike | None, list[PathLike]]:
    """Finds the domain and utility_function files in a folder"""
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    folder_name = str(folder_name)
    for f in files:
        if not f.endswith(".json") or f.endswith("specials.json"):
            continue
        full_name = folder_name + "/" + f
        data = load(full_name)

        if data and "issuesValues" in data.keys():
            domain_file_name = full_name
        elif data and "LinearAdditiveUtilitySpace" in data.keys():
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names


def find_domain_and_utility_files_xml(
    folder_name,
) -> tuple[PathLike | None, list[PathLike]]:
    """Finds the domain and utility_function files in a folder"""
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    folder_name = str(folder_name)
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names  #


def load_geniusweb_domain_from_folder(
    folder_name: str | PathLike,
    ignore_reserved=False,
    ignore_discount=False,
    use_reserved_outcome=False,
    safe_parsing=False,
    **kwargs,
) -> Scenario:
    """
    Loads a genius-web domain from a folder. See ``load_geniusweb_domain`` for more details.

    Args:
        folder_name: A folder containing one XML domain file and one or more ufun files in Genius format
        ignore_reserved: Sets the reserved_value of all ufuns to -inf
        ignore_discount: Ignores discounting
        safe_parsing: Applies more stringent checks during parsing
        kwargs: Extra arguments to pass verbatim to SAOMechanism constructor

    Returns:
        A domain ready for `make_session`

    """
    folder_name = str(folder_name)
    domain_file_name, utility_file_names = find_geniusweb_domain_and_utility_files(
        folder_name
    )

    if domain_file_name is None:
        raise ValueError("Cannot find a domain file")
    return load_geniusweb_domain(
        domain_file_name=domain_file_name,
        utility_file_names=utility_file_names,
        safe_parsing=safe_parsing,
        ignore_reserved=ignore_reserved,
        ignore_discount=ignore_discount,
        use_reserved_outcome=use_reserved_outcome,
        **kwargs,
    )


def find_geniusweb_domain_and_utility_files(
    folder_name,
) -> tuple[PathLike | None, list[PathLike]]:
    """Finds the domain and utility_function files in a GeniusWeb formatted json folder"""
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    folder_name = str(folder_name)
    for f in files:
        if not f.endswith(".json"):
            continue
        full_name = folder_name + "/" + f
        d = load(full_name)
        if any(_ in d.keys() for _ in GENIUSWEB_UFUN_TYPES):
            utility_file_names.append(full_name)
        elif "issuesValues" in d.keys():
            domain_file_name = full_name
    return domain_file_name, utility_file_names


def find_genius_domain_and_utility_files(
    folder_name,
) -> tuple[PathLike | None, list[PathLike]]:
    """Finds the domain and utility_function files in a folder"""
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    folder_name = str(folder_name)
    for f in files:
        if not f.endswith(".xml") or f.endswith("pareto.xml"):
            continue
        full_name = folder_name + "/" + f
        root = ET.parse(full_name).getroot()

        if root.tag == "negotiation_template":
            domain_file_name = full_name
        elif root.tag == "utility_space":
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names  #


def load_geniusweb_domain(
    domain_file_name: PathLike,
    utility_file_names: Iterable[PathLike] | None = None,
    ignore_discount=False,
    ignore_reserved=False,
    use_reserved_outcome=False,
    safe_parsing=True,
    **kwargs,
) -> Scenario:
    """
    Loads a geniusweb domain, creates appropriate negotiators if necessary

    Args:
        domain_file_name: XML file containing Genius-formatted domain spec
        utility_file_names: XML files containing Genius-fromatted ufun spec
        ignore_reserved: Sets the reserved_value of all ufuns to -inf
        ignore_discount: Ignores discounting
        safe_parsing: Applies more stringent checks during parsing
        kwargs: Extra arguments to pass verbatim to SAOMechanism constructor

    Returns:
        A `Domain` ready to run

    """

    issues = None
    name = load(domain_file_name).get("name", domain_file_name)
    if domain_file_name is not None:
        issues, _ = issues_from_geniusweb(domain_file_name, safe_parsing=safe_parsing)

    agent_info = []
    if utility_file_names is None:
        utility_file_names = []
    for ufname in utility_file_names:
        utility, discount_factor = UtilityFunction.from_geniusweb(
            file_name=ufname,
            issues=issues,
            safe_parsing=safe_parsing,
            ignore_discount=ignore_discount,
            ignore_reserved=ignore_reserved,
            use_reserved_outcome=use_reserved_outcome,
            name=str(ufname),
        )
        agent_info.append(
            {
                "ufun": utility,
                "ufun_name": ufname,
                "reserved_value_func": utility.reserved_value
                if utility is not None
                else float("-inf"),
                "discount_factor": discount_factor,
            }
        )
    if domain_file_name is not None:
        kwargs["dynamic_entry"] = False
        kwargs["max_n_agents"] = None
        if not ignore_discount:
            for info in agent_info:
                info["ufun"] = (
                    info["ufun"]
                    if info["discount_factor"] is None or info["discount_factor"] == 1.0
                    else make_discounted_ufun(
                        ufun=info["ufun"],
                        discount_per_round=info["discount_factor"],
                        power_per_round=1.0,
                    )
                )
    if issues is None:
        raise ValueError(f"Could not load domain {domain_file_name}")

    return Scenario(
        outcome_space=make_os(issues, name=name),
        ufuns=[_["ufun"] for _ in agent_info],  # type: ignore We trust that the ufun will be loaded
    )



from __future__ import annotations
from typing import Callable, Mapping, Union

from negmas.common import Value
from negmas.helpers.prob import Distribution
from negmas.outcomes.common import Outcome

__all__ = ["INVALID_UTILITY", "Distribution", "Value", "UtilityValue"]

INVALID_UTILITY = float("-inf")

UtilityValue = Value
"""
An alias for `Value` which is either a probability `Distribution` or a `float`
"""

OutcomeUtilityMapping = Union[
    Callable[[Outcome | None], Value], Mapping[Outcome | None, Value]
]
"""A mapping from an outcome to its utility value"""



from __future__ import annotations
import pytest

from negmas.gb.negotiators.timebased import AspirationNegotiator
from negmas.outcomes import make_issue, make_os
from negmas.plots.util import plot_offline_run
from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction
from negmas.sao.mechanism import SAOMechanism


@pytest.mark.skip("A test added for manually testing plotting")
def test_mechanism_plot():
    os = make_os([make_issue(10) for _ in range(3)])
    m = SAOMechanism(n_steps=100, outcome_space=os)
    ufuns = LinearAdditiveUtilityFunction.generate_random_bilateral(
        list(os.enumerate_or_sample())
    )
    for i, u in enumerate(ufuns):
        u.outcome_space = os
        m.add(AspirationNegotiator(id=f"a{i}", name=f"n{i}", ufun=u))
    m.run()
    m.plot()
    import matplotlib.pyplot as plt

    plt.show()


@pytest.mark.skip("A test added for manually testing plotting")
def test_offline_plot():
    os = make_os([make_issue(10) for _ in range(3)])
    m = SAOMechanism(n_steps=100, outcome_space=os)
    ufuns = LinearAdditiveUtilityFunction.generate_random_bilateral(
        list(os.enumerate_or_sample())
    )
    for i, u in enumerate(ufuns):
        u.outcome_space = os
        m.add(AspirationNegotiator(id=f"a{i}", name=f"n{i}", ufun=u))
    m.run()
    trace = m.full_trace
    ids = m.negotiator_ids
    names = m.negotiator_names
    state = m.state

    plot_offline_run(
        trace,
        ids,
        ufuns,
        state.agreement,
        state.timedout,
        state.broken,
        state.has_error,
        names=names,
    )
    import matplotlib.pyplot as plt

    plt.show()



from __future__ import annotations
import time

from ..common import MechanismState, Value
from ..outcomes import Outcome
from .base import BaseElicitor

__all__ = ["DummyElicitor", "FullKnowledgeElicitor"]


class DummyElicitor(BaseElicitor):
    """
    A dummy elicitation algorithm that does not do any elicitation.
    """

    def utility_on_rejection(self, outcome: Outcome, state: MechanismState) -> Value:
        return self.reserved_value

    def can_elicit(self) -> bool:
        return True

    def elicit_single(self, state: MechanismState):
        return False

    def init_elicitation(
        self, preferences: IPUtilityFunction | Distribution | None, **kwargs
    ):
        super().init_elicitation(preferences=preferences, **kwargs)
        strt_time = time.perf_counter()
        self.offerable_outcomes = self._nmi.outcomes
        self._elicitation_time += time.perf_counter() - strt_time


class FullKnowledgeElicitor(BaseElicitor):
    """
    An elicitor that does not *need* to do any elicitation because it has full access
    to the user ufun.
    """

    def utility_on_rejection(self, outcome: Outcome, state: MechanismState) -> Value:
        return self.reserved_value

    def can_elicit(self) -> bool:
        return True

    def elicit_single(self, state: MechanismState):
        return False

    def init_elicitation(
        self, preferences: IPUtilityFunction | Distribution | None, **kwargs
    ):
        super().init_elicitation(preferences=self.user.ufun)
        strt_time = time.perf_counter()
        self.offerable_outcomes = self._nmi.outcomes
        self._elicitation_time += time.perf_counter() - strt_time



"""
Helper modueles
"""

from __future__ import annotations

from .types import *

# from .inout import *
# from .prob import *
# from .numeric import *
from .strings import *
from .logging import *
from .timeout import *
from .misc import *
from .numba_checks import *

__all__ = (
    types.__all__
    + strings.__all__
    + logging.__all__
    + timeout.__all__
    + misc.__all__
    + numba_checks.__all__
    # + ["prob", "inout", "numeric"]
    # + numeric.__all__
    # + prob.__all__
    # + inout.__all__
)



# -*- coding: utf-8 -*-
""""
The interface to all negotiators capable of eliciting user preferences before
, and during negotiations.
"""

from __future__ import annotations

from .base import *
from .baseline import *
from .common import *
from .mechanism import *
from .pandora import *
from .queries import *
from .strategy import *
from .user import *
from .voi import *

__all__ = (
    common.__all__
    + strategy.__all__
    + user.__all__
    + queries.__all__
    + base.__all__
    + baseline.__all__
    + pandora.__all__
    + voi.__all__
    + mechanism.__all__
)

import numpy as np

np.seterr(all="raise")  # setting numpy to raise exceptions in case of errors



from __future__ import annotations
import itertools
import random
from typing import Any, Callable, Iterable, Literal, overload

import numpy as np

from negmas.helpers.misc import (
    distribute_integer_randomly,
    floatin,
    generate_random_weights,
    intin,
)
from negmas.negotiators.helpers import PolyAspiration, TimeCurve
from negmas.outcomes import make_issue, make_os
from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction
from negmas.preferences.crisp.mapping import MappingUtilityFunction
from negmas.preferences.value_fun import TableFun

__all__ = [
    "sample_between",
    "make_pareto",
    "make_endpoints",
    "make_non_pareto",
    "generate_utility_values",
    "generate_multi_issue_ufuns",
    "generate_single_issue_ufuns",
    "ParetoGenerator",
    "GENERATOR_MAP",
    "make_curve_pareto",
    "make_piecewise_linear_pareto",
    "make_zero_sum_pareto",
]


def make_curve(
    curve: TimeCurve
    | Literal["boulware"]
    | Literal["conceder"]
    | Literal["linear"]
    | float,
    starting_utility: float = 1.0,
) -> TimeCurve:
    """
    Generates a `TimeCurve` or `Aspiration` with optional `starting_utility`self.

    Default behavior is to return a `PolyAspiration` object.
    """
    if isinstance(curve, TimeCurve):
        return curve
    return PolyAspiration(starting_utility, curve)


def sample_between(
    start: float,
    end: float,
    n: int,
    endpoint=True,
    main_range: tuple[float, float] = (0.3, 0.5),
):
    """
    Samples n values between the start and end given, optionally with the endpoint included.

    Remarks:
        - The samples are drawn to be somewhere between main_range limits of the range between start and end.
        - Samples that end up smaller than start are sent back to the first 50% of the range between start and end.
        - Samples that end up larger than end are sent back to the first 50% of the range between start and end.
    """
    if n == 0:
        return []
    if end - start < 1e-8:
        return [start] * n
    samples = np.linspace(start, end, num=n, endpoint=endpoint)
    if len(samples) > 2:
        samples[1:-1] += (
            main_range[0]
            * (end - start)
            * (np.random.random(len(samples) - 2) - main_range[1])
        )
        samples[samples < start] = start + np.random.random(
            len(samples[samples < start])
        ) * 0.5 * (end - start)
        samples[samples > end] = end - np.random.random(
            len(samples[samples > end])
        ) * 0.5 * (end - start)
    return samples


def make_endpoints(
    n_segments: int | tuple[int, int] | list[int],
    xrange: tuple[tuple[float, float], tuple[float, float]] = ((0, 1), (0, 1)),
) -> list[tuple[float, ...]]:
    """Create endpoints for n_segments with the ranges given"""
    n = intin(n_segments) + 1
    x = np.sort(sample_between(*xrange[0], n))
    y = np.sort(sample_between(*xrange[1], n))[::-1]
    return list(zip(x, y))


def make_pareto(
    endpoints: list[tuple[float, ...]], n_outcomes: int
) -> list[tuple[float, ...]]:
    """
    Generates a piecewise linear curve with a specified number of segments and all slopes negative.

    Args:
      endpoints: A list of (x, y) coordinates for the endpoints of each segment.
      n_outcomes: number of values to distribute in the piecewise linear lines connecting the endpoints

    Returns:
        list of tuples representing utilities at the pareto-frontier

    Remarks:
      - If the number of end points is less than n_outcomes + 1,
    """
    if not endpoints and n_outcomes < 1:
        return []
    if len(endpoints) < 2:
        return [endpoints[0]] * n_outcomes
    if n_outcomes < 2:
        return [endpoints[0]] * n_outcomes
    num_segments = len(endpoints) - 1
    if n_outcomes <= num_segments:
        n_per_segment = [1] * (n_outcomes - 1) + [0] * (num_segments - n_outcomes) + [1]
    else:
        n_per_segment = distribute_integer_randomly(n_outcomes - 1, num_segments)

    # Calculate slopes based on difference in y-coordinates
    slopes = [
        (endpoints[i + 1][1] - endpoints[i][1])
        / (endpoints[i + 1][0] - endpoints[i][0])
        for i in range(num_segments)
    ]

    points = []

    # Iterate through each segment and generate its points
    for i in range(num_segments):
        start_x, start_y = endpoints[i]
        end_x = endpoints[i + 1][0]
        slope = slopes[i]
        n = n_per_segment[i] - 1
        if n == 0:
            end_y = endpoints[i + 1][1]
            if i == 0:
                points.append((start_x, start_y))
            elif i < num_segments - 1:
                r = random.random()
                if r < 0.5:
                    points.append((start_x, start_y))
                else:
                    points.append((end_x, end_y))
            else:
                points.append((end_x, end_y))
            continue
        if n < 0:
            continue
        samples = sample_between(start_x, end_x, n, endpoint=False)
        points.append((start_x, start_y))

        # Calculate x and y values for each point on the segment
        for x in samples:
            y = start_y + slope * (x - start_x)
            points.append((x, y))
    # append the last point representing the last endpoint
    if len(points) < n_outcomes:
        points.append((endpoints[-1][0], endpoints[-1][1]))
    return points


def make_non_pareto(
    pareto: list[tuple[float, ...]],
    n: int,
    limit: tuple[float, float] = (0, 0),
    eps: float = 1e-4,
) -> list[tuple[float, ...]]:
    """Adds extra outcomes dominated by the given Pareto outcomes

    Args:
        pareto: list of pareto points in the utility space.
        n: number of outcomes to add that are dominated by the pareto outcomes
        limit: The minimum x and y values allowed
        eps: A small margin
    """
    if n < 1:
        return []
    if len(pareto) < 1:
        raise ValueError(
            "Cannot use a pareto frontier with zero points. No such thing can exist!!"
        )
    ndims = len(pareto[0])
    available_indices, ok = [], []
    # remove points in the pareto under the limit from consideration
    for p in pareto:
        available = []
        for d in range(ndims):
            if p[d] < limit[d] + eps:
                continue
            available.append(d)
            available_indices.append(tuple(available))
        if not available_indices:
            continue
        ok.append(len(available_indices[-1]) > 0)
    _ok = np.asarray(ok, dtype=bool)
    pareto = np.asarray(pareto)[_ok].tolist()
    available_indices = [_ for _, ok in zip(available_indices, _ok) if ok]
    n_pareto = len(pareto)
    if n_pareto < 1:
        return []
    if n_pareto == 1:
        points = np.zeros(n, dtype=int)
    else:
        points = np.random.randint(0, n_pareto - 1, size=n, dtype=int)
    extra = []
    if len(points) < 1:
        return []
    for i in points:
        p, available = pareto[i], available_indices[i]
        new = [_ for _ in p]
        if len(available) == 1:
            changes = available
        else:
            changes = random.sample(available, k=random.randint(1, len(available)))
        for d in changes:
            new[d] -= random.random() * (p[d] - limit[d] - eps) - eps
        extra.append(tuple(new))

    return extra


def make_zero_sum_pareto(n_pareto: int):
    """Generate a zero-sum Pareto generator of the given number of segments"""
    return make_piecewise_linear_pareto(n_pareto, n_segments=1)


def make_piecewise_linear_pareto(
    n_pareto: int, *, n_segments: int | list[int] | tuple[int, int] = (2, 5)
):
    """Generate a piecewise-linear Pareto generator of the given number of segments"""
    endpoints = make_endpoints(n_segments)
    if len(endpoints) < n_pareto:
        endpoints = endpoints[: n_pareto - 1]
    return make_pareto(endpoints, n_pareto)


def make_curve_pareto(
    n_pareto: int,
    *,
    shape: TimeCurve
    | Literal["boulware"]
    | Literal["conceder"]
    | Literal["linear"]
    | float
    | tuple[float, float]
    | list[float] = (0.1, 6),
):
    """Generate a Pareto curve of the given shape."""
    if isinstance(shape, Iterable) and not isinstance(shape, str):
        shape = floatin(shape, log_uniform=True)
    curve = make_curve(shape)
    x = 1.0 - np.sort(np.random.random(n_pareto), axis=None)
    y = [curve.utility_range(_) for _ in x]
    y = [(a + b) / 2 for a, b in y]
    return list(zip(x, y, strict=True))


ParetoGenerator = Callable[[int], list[tuple[float, ...]]]
"""Type of Pareto generators. Receives a number of points and returns a list of utility values corresponding to a Pareto front of that size"""

GENERATOR_MAP: dict[str, ParetoGenerator] = dict(
    piecewise_linear=make_piecewise_linear_pareto,
    curve=make_curve_pareto,
    zero_sum=make_zero_sum_pareto,
)


def generate_utility_values(
    n_pareto: int,
    n_outcomes: int,
    n_ufuns: int = 2,
    pareto_first=False,
    pareto_generator: ParetoGenerator | str = "piecewise_linear",
    generator_params: dict[str, Any] | None = None,
) -> list[tuple[float, ...]]:
    """
    Generates ufuns that have a controllable Pareto frontier

    Args:
        n_pareto: Number of outcomes on the Pareto frontier
        n_outcomes: Total number of outcomes to generate
        pareto_first: If given, Pareto outcomes come first in the returned results.
        pareto_generator: The method used to generate the Pareto front utilities
        generator_params: The parameters passed to the Pareto generator

    Returns:
        A list of tuples each giving the utilities of one outcome
    """
    if n_ufuns != 2:
        raise NotImplementedError(
            f"We only support generation of two ufuns using this method. {n_ufuns} ufuns are requested."
        )
    if not generator_params:
        generator_params = dict()
    if isinstance(pareto_generator, str):
        pareto_generator = GENERATOR_MAP[pareto_generator]
    pareto = pareto_generator(n_pareto, **generator_params)
    extra = make_non_pareto(pareto, n_outcomes - n_pareto)
    # print(len(xvals), len(extra), n_segments, n_pareto, n_outcomes)
    points = list(pareto) + extra
    if not pareto_first:
        random.shuffle(points)
    return points


def zip_cycle(A, *args):
    """Zips generators A and B, cycling through B as needed if it's shorter.

    Args:
        A: The first generator.
        args: Other generators

    Yields:
        Pairs of elements from A and B, cycling through B as needed.

    Raises:
        ValueError: If B is longer than A.
    """

    if len(A) < min(len(_) for _ in args):  # Check for invalid length relationship
        raise ValueError("Generator A must be equal to or longer than generator B.")

    cyc = (itertools.cycle(_) for _ in args)
    return zip(A, *cyc)


def zip_cycle_longest(A, B):
    """Zips generators A and B, cycling through B as needed if it's shorter.

    Args:
        A: The first generator.
        B: The second generator.

    Yields:
        Pairs of elements from A and B, cycling through B as needed.

    Raises:
        ValueError: If B is longer than A.
    """

    if len(A) < len(B):  # Check for invalid length relationship
        raise ValueError("Generator A must be equal to or longer than generator B.")

    B_cycle = itertools.cycle(B)  # Create a cycle of the shorter generator B
    return zip(A, B_cycle)  # Zip A with the cycled B


def _adjust_ufuns(
    ufuns,
    os,
    linear,
    rational_fractions,
    selector: Callable[[float, float], float] = max,
):
    outcomes = None
    if rational_fractions:
        for u, f in zip_cycle_longest(ufuns, rational_fractions):
            outcomes = os.enumerate_or_sample()
            vals = sorted(u(_) for _ in outcomes)
            n_outcomes = len(vals)
            limit = max(0, min(len(vals) - 1, int((1 - f) * len(vals) + 0.5)))
            if f * n_outcomes <= 1:
                r = vals[-1] + 0.001
            elif f >= 1.0:
                r = vals[0] - 0.001
            else:
                r = vals[limit] - 1e-9
            u.reserved_value = selector(r, u.reserved_value)

    if linear:
        return ufuns
    maps = []
    if outcomes is None:
        outcomes = os.enumerate_or_sample()
    for u in ufuns:
        maps.append(
            MappingUtilityFunction(
                dict(zip(outcomes, [u(_) for _ in outcomes])),
                outcome_space=os,
                name=u.name,
                reserved_value=u.reserved_value,
            )
        )
    return tuple(maps)


def generate_single_issue_ufuns(
    n_pareto: int,
    n_outcomes: int,
    n_ufuns: int = 2,
    pareto_first=False,
    pareto_generator: ParetoGenerator | str = "piecewise_linear",
    generator_params: dict[str, Any] | None = None,
    reserved_values: list[float] | tuple[float, float] | float = 0.0,
    rational_fractions: list[float] | None = None,
    reservation_selector: Callable[[float, float], float] = max,
    issue_name: str = "portions",
    os_name: str = "S",
    ufun_names: tuple[str, ...] | list[str] | None = None,
    numeric: bool = False,
    linear: bool = True,
) -> tuple[LinearAdditiveUtilityFunction | MappingUtilityFunction, ...]:
    """Generates a set of single-issue ufuns

    Args:
        pareto_first: return the pareto outcomes first in the outcome-space
        n_pareto: number of pareto outcomes
        n_outcomes: number of outcomes. Must be >= `n_pareto`
        n_ufuns: number of ufuns
        pareto_generator: The generation method. See `GENERATOR_MAP`
        generator_params: parameters of the generator
        reserved_values: Reserved values to use for generated ufuns.
                         A list to cycle through, or a tuple to sample within or a single value to repeat.
        rational_fractions: Fraction of rational outcomes for each ufun. Should have `n_ufuns` values
                            (or it will cycle). If given with `reserved_values`, the `reservation_selector`
                            will be used to select the final reservation value
        reservation_selector: A function that receives the reserved value suggested by rational_fraction and
                               that suggested by reserved_value and selects the final reserved values. The
                               default is `max`. You can replace that with `min`, the first, the second,
                               mean, or any combination.
        issue_name: Name of the single issue.
        os_name: Name of the outcomes space in the ufuns created
        ufun_names: Names of utility functions
        numeric: All issue values are numeric (integers)
        linear: Whether to use a linear-additive-ufun instead of the general mapping ufun

    Returns:
        A tuple of single-issue ufuns
    """
    if (
        isinstance(reserved_values, tuple)
        and len(reserved_values) == 2
        and reserved_values[0] <= reserved_values[1]
    ):
        reserved_values = [
            floatin(reserved_values, log_uniform=False) for _ in range(n_ufuns)
        ]
    elif not isinstance(reserved_values, Iterable):
        reserved_values = [float(reserved_values)] * n_ufuns
    vals = generate_utility_values(
        n_pareto, n_outcomes, n_ufuns, pareto_first, pareto_generator, generator_params
    )
    n = len(vals)

    if numeric:
        issues = (make_issue(n, issue_name),)
    else:
        issues = (make_issue([f"{i}_{n-1 - i}" for i in range(n)], issue_name),)
    os = make_os(issues, name=os_name)
    if not ufun_names:
        ufun_names = tuple(f"u{i}" for i in range(n_ufuns))
    ufuns = tuple(
        LinearAdditiveUtilityFunction(
            values=(
                TableFun({_: float(vals[i][k]) for i, _ in enumerate(issues[0].all)}),
            ),
            name=uname,
            outcome_space=os,
            reserved_value=r,
        )
        for k, (r, uname) in enumerate(zip_cycle_longest(ufun_names, reserved_values))
        # for k, (uname, r) in enumerate(zip(("First", "Second"), reserved_ranges))
    )
    return _adjust_ufuns(ufuns, os, linear, rational_fractions, reservation_selector)


@overload
def generate_multi_issue_ufuns(
    n_issues: int,
    n_values: int | tuple[int, int] = 0,
    sizes: None = None,
    n_ufuns: int = 2,
    pareto_generators: tuple[ParetoGenerator | str, ...] = ("piecewise_linear",),
    generator_params: tuple[dict[str, Any], ...] | None = None,
    reserved_values: list[float] | tuple[float, float] | float = 0.0,
    rational_fractions: list[float] | None = None,
    reservation_selector: Callable[[float, float], float] = max,
    issue_names: tuple[str, ...] | list[str] | None = None,
    os_name: str | None = None,
    ufun_names: tuple[str, ...] | None = None,
    numeric: bool = False,
    linear: bool = True,
) -> tuple[LinearAdditiveUtilityFunction | MappingUtilityFunction, ...]:
    ...


@overload
def generate_multi_issue_ufuns(
    n_issues: int,
    n_values: None = None,
    sizes: tuple[int, ...] | list[int] = tuple(),
    n_ufuns: int = 2,
    pareto_generators: tuple[ParetoGenerator | str, ...] = ("piecewise_linear",),
    generator_params: tuple[dict[str, Any], ...] | None = None,
    reserved_values: list[float] | tuple[float, float] | float = 0.0,
    rational_fractions: list[float] | None = None,
    reservation_selector: Callable[[float, float], float] = max,
    issue_names: tuple[str, ...] | list[str] | None = None,
    os_name: str | None = None,
    ufun_names: tuple[str, ...] | None = None,
    numeric: bool = False,
    linear: bool = True,
) -> tuple[LinearAdditiveUtilityFunction | MappingUtilityFunction, ...]:
    ...


def generate_multi_issue_ufuns(
    n_issues: int,
    n_values: int | tuple[int, int] | None = None,
    sizes: None | tuple[int, ...] | list[int] = None,
    n_ufuns: int = 2,
    pareto_generators: tuple[ParetoGenerator | str, ...] = tuple(GENERATOR_MAP.keys()),
    generator_params: tuple[dict[str, Any], ...] | None = None,
    reserved_values: list[float] | tuple[float, float] | float = 0.0,
    rational_fractions: list[float] | None = None,
    reservation_selector: Callable[[float, float], float] = max,
    issue_names: tuple[str, ...] | list[str] | None = None,
    os_name: str | None = None,
    ufun_names: tuple[str, ...] | None = None,
    numeric: bool = False,
    linear: bool = True,
) -> tuple[LinearAdditiveUtilityFunction | MappingUtilityFunction, ...]:
    """Generates a set of ufuns with an outcome space of the given number of issues.

    Args:
        n_issues: Number of issues
        n_values: Range of issue sizes. Only used if `sizes` is not passed
        sizes: Sizes of issues
        n_ufuns: Number of ufuns to generate
        pareto_generators: The generators to use internally to generate value functions
        generator_params: parameters of the generators.
        reserved_values: Reserved values to use for generated ufuns.
                         A list to cycle through, or a tuple to sample within or a single value to repeat.
        rational_fractions: Fraction of rational outcomes for each ufun. Should have `n_ufuns` values
                            (or it will cycle). If given with `reserved_values`, the `reservation_selector`
                            will be used to select the final reservation value
        reservation_selector: A function that receives the reserved value suggested by rational_fraction and
                               that suggested by reserved_value and selects the final reserved values. The
                               default is `max`. You can replace that with `min`, the first, the second,
                               mean, or any combination.
        os_name: Name of the outcomes space in the ufuns created
        ufun_names: Names of utility functions
        numeric: All issue values are numeric (integers)
        linear: Whether to use a linear-additive-ufun instead of the general mapping ufun

    Returns:
        A tuple of `n_ufuns` utility functions.
    """
    if (
        isinstance(reserved_values, tuple)
        and len(reserved_values) == 2
        and reserved_values[0] <= reserved_values[1]
    ):
        reserved_values = [
            floatin(reserved_values, log_uniform=False) for _ in range(n_ufuns)
        ]
    elif not isinstance(reserved_values, Iterable):
        reserved_values = [float(reserved_values)] * n_ufuns
    if ufun_names is None:
        ufun_names = tuple(f"u{i+1}" for i in range(n_ufuns))
    vals = [dict() for _ in range(n_ufuns)]
    if not generator_params:
        generator_params = [dict() for _ in pareto_generators]  # type: ignore
    if not issue_names:
        issue_names = tuple(f"i{k+1}" for k in range(n_issues))
    gp = list(zip(pareto_generators, generator_params))  # type: ignore
    if sizes is None:
        assert n_values is not None
        sizes = tuple(intin(n_values) for _ in range(n_issues))  # type: ignore
    assert sizes is not None
    for i, n in enumerate(sizes):
        g, p = random.choice(gp)
        v = generate_utility_values(
            n, n, n_ufuns, pareto_generator=g, generator_params=p
        )
        for j in range(n_ufuns):
            vals[j][i] = [float(_[j]) for _ in v]
    weights = [float(_) for _ in generate_random_weights(n_issues)]
    os = make_os(
        issues=[
            make_issue(ni, name=iname)
            if numeric
            else make_issue([f"v{k+1}" for k in range(ni)], name=f"i{i+1}")
            for i, (ni, iname) in enumerate(zip(sizes, issue_names))
        ],
        name=os_name,
    )
    ufuns = tuple(
        LinearAdditiveUtilityFunction(
            values=[
                TableFun(
                    dict(
                        zip(
                            [
                                k if numeric else f"v{k+1}"
                                for k in range(len(vals[j][i]))
                            ],
                            vals[j][i],
                        )
                    )
                )
                for i in range(n_issues)
            ],
            weights=weights,
            outcome_space=os,
            name=name,
            reserved_value=r,
        )
        for j, (name, r) in enumerate(zip_cycle_longest(ufun_names, reserved_values))
    )
    return _adjust_ufuns(ufuns, os, linear, rational_fractions, reservation_selector)


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from rich import print

    fig, subplots = plt.subplots(4, 4, sharex=True, sharey=True, squeeze=True)
    subplots = subplots.flatten()
    choices = list(GENERATOR_MAP.keys())

    for ax in subplots:
        if random.random() > 0.33:
            n_pareto = random.randint(30, 100)
            n_outcomes = 400
        else:
            n_pareto = n_outcomes = 400
        method = random.choice(choices)
        if method == "zero_sum":
            choices = [_ for _ in choices if _ != "zero_sum"]
        params = dict(
            piecewise_linear=dict(n_segments=random.randint(1, min(n_pareto, 20))),
            curve=dict(
                shape=4
                * round(
                    random.random() if random.random() > 0.5 else 1 / random.random(), 2
                )
            ),
        ).get(method, dict())
        points = generate_utility_values(
            n_pareto,
            n_outcomes,
            pareto_first=True,
            generator_params=params,
            pareto_generator=method,
        )
        assert len(points) == n_outcomes, f"{len(points)=}, {n_outcomes=}"
        print(
            f"Generated {len(set(points))=} unique outcomes for: {n_outcomes=}, {n_pareto=}, {method=}, {params=}"
        )
        ax.scatter([_[0] for _ in points], [_[1] for _ in points], color="b")
        ax.plot(
            [_[0] for _ in points[:n_pareto]],
            [_[1] for _ in points[:n_pareto]],
            color="r",
            marker="x",
        )
        ax.set_title(f"{method}: {str(params)}")
    plt.show()



from __future__ import annotations
#
#
# import random
# import time
# from pathlib import Path
# from typing import Iterable
#
# import hypothesis.strategies as st
# import pkg_resources
# import pytest
# from hypothesis import Verbosity, example, given, settings
# from matplotlib.axes import itertools
#
# from negmas.gb.evaluators.tau import INFINITE
# from negmas.gb.mechanisms.mechanisms import (
#     GAOMechanism,
#     GeneralizedTAUMechanism,
#     TAUMechanism,
# )
# from negmas.gb.negotiators.cab import CABNegotiator, CANNegotiator, CARNegotiator
# from negmas.gb.negotiators.timebased import AspirationNegotiator
# from negmas.gb.negotiators.war import WABNegotiator, WANNegotiator, WARNegotiator
# from negmas.inout import Scenario
# from negmas.mechanisms import Mechanism
# from negmas.outcomes.base_issue import make_issue
# from negmas.outcomes.categorical_issue import CategoricalIssue
# from negmas.outcomes.common import Outcome
# from negmas.outcomes.outcome_space import DiscreteCartesianOutcomeSpace, make_os
# from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction as LU
# from negmas.preferences.crisp.mapping import MappingUtilityFunction
# from negmas.preferences.crisp_ufun import UtilityFunction
# from negmas.preferences.ops import dominating_points, pareto_frontier
# from negmas.preferences.value_fun import AffineFun, IdentityFun, LinearFun, TableFun
# from negmas.sao.mechanism import SAOMechanism
# from tests.switches import NEGMAS_FASTRUN, NEGMAS_RUN_GENIUS
#
# SHOW_PLOTS = False
# SHOW_ALL_PLOTS = False
# FORCE_PLOT = False
# SHOW_HISTORY = False
# SHOW_ALL_HISTORIES = False
# FORCE_HISTORY = False
# SAONEGOTIATORS = [
#     AspirationNegotiator,
# ]
# MECHS = (TAUMechanism,) if NEGMAS_FASTRUN else (TAUMechanism, GeneralizedTAUMechanism)
# # TODO: resolve the infinite loop in WAN and WAB
# NEGOTIATORS = [
#     WARNegotiator,
#     CABNegotiator,
#     CANNegotiator,
#     CARNegotiator,
#     WANNegotiator,
#     WABNegotiator,
# ]
#
# PROPOSED = [
#     WARNegotiator,
#     CABNegotiator,
# ]
#
# RATIONAL_REJECTOR = [
#     CANNegotiator,
#     WANNegotiator,
#     CABNegotiator,
#     WABNegotiator,
# ]
#
# BEST_ACCEPTOR = [
#     CABNegotiator,
#     WABNegotiator,
# ]
# # all proposed TAU strategies are rational
# RATIONAL = list(itertools.product(NEGOTIATORS, NEGOTIATORS))
# IRRATIONAL = [
#     _ for _ in itertools.product(NEGOTIATORS, NEGOTIATORS) if _ not in RATIONAL
# ]
#
#
# # If one side is WAR and the other side may reject a rational outcome, the negotiation may be incomplete
# INCOMPLETE = list(
#     set(
#         list(
#             itertools.chain(
#                 *(
#                     [
#                         itertools.permutations(_)
#                         for _ in zip(itertools.repeat(WANNegotiator), RATIONAL_REJECTOR)
#                     ]
#                     + [
#                         itertools.permutations(_)
#                         for _ in zip(itertools.repeat(WARNegotiator), RATIONAL_REJECTOR)
#                     ]
#                     + [
#                         itertools.permutations(_)
#                         for _ in zip(itertools.repeat(WABNegotiator), RATIONAL_REJECTOR)
#                     ]
#                 )
#             )
#         )
#     )
# )
# COMPLETE = [
#     _ for _ in itertools.product(NEGOTIATORS, NEGOTIATORS) if _ not in INCOMPLETE
# ]
#
# # An optimal strategy profile is one for which both sides are from the proposed set
# OPTIMAL = list(itertools.product(BEST_ACCEPTOR, BEST_ACCEPTOR))
# NONOPTIMAL = [
#     _ for _ in itertools.product(NEGOTIATORS, NEGOTIATORS) if _ not in OPTIMAL
# ]
#
#
# EQUILIBRIUM = [
#     (WARNegotiator, WARNegotiator),
# ]
# NONEQUILIBRIUM = [
#     _ for _ in itertools.product(NEGOTIATORS, NEGOTIATORS) if _ not in EQUILIBRIUM
# ]
#
#
# NORAISE = [
#     (WARNegotiator, WARNegotiator),
#     (WARNegotiator, CANNegotiator),
#     (WARNegotiator, WANNegotiator),
#     (WARNegotiator, CABNegotiator),
#     (WARNegotiator, WABNegotiator),
#     (CANNegotiator, WARNegotiator),
#     (CANNegotiator, WANNegotiator),
#     (CANNegotiator, WABNegotiator),
#     (WANNegotiator, WARNegotiator),
#     (WANNegotiator, CANNegotiator),
#     (WANNegotiator, WANNegotiator),
#     (WANNegotiator, CABNegotiator),
#     (WANNegotiator, WABNegotiator),
#     (CABNegotiator, WARNegotiator),
#     (CABNegotiator, WANNegotiator),
#     (CABNegotiator, WABNegotiator),
#     (WABNegotiator, WARNegotiator),
#     (WABNegotiator, CANNegotiator),
#     (WABNegotiator, WANNegotiator),
#     (WABNegotiator, CABNegotiator),
#     (WABNegotiator, WABNegotiator),
# ]
#
#
# @pytest.fixture
# def scenarios_folder():
#     return pkg_resources.resource_filename(
#         "negmas", resource_name="tests/data/scenarios"
#     )
#
#
# def _plot(p, err=False, force=False):
#     if not force and err and not SHOW_PLOTS:
#         return
#     if not force and not err and not SHOW_ALL_PLOTS:
#         return
#     import matplotlib.pyplot as plt
#
#     p.plot(xdim="step")
#     plt.show()
#     plt.savefig("fig.png")
#
#
# def _history(p: Mechanism, err=False, force=False):
#     if not force and err and not SHOW_HISTORY:
#         return ""
#     if not force and not err and not SHOW_ALL_HISTORIES:
#         return ""
#     return p.trace  # type: ignore
#
#
# @pytest.mark.parametrize("neg", NEGOTIATORS)
# def test_a_tau_session_example(neg):
#     for _ in range(100):
#         r1, r2, n1, n2 = 0.2, 0.3, 5, 10
#         eps = 1e-3
#         time.perf_counter()
#         os: DiscreteCartesianOutcomeSpace = make_os([make_issue(n1), make_issue(n2)])  # type: ignore
#         p = TAUMechanism(outcome_space=os)
#         ufuns = [LU.random(os, reserved_value=r1), LU.random(os, reserved_value=r2)]
#         for i, u in enumerate(ufuns):
#             p.add(
#                 neg(name=f"RCS{i}", id=f"RCS{i}"),
#                 preferences=u,
#             )
#         front_utils, front_outcomes = p.pareto_frontier()
#         no_valid_outcomes = all(
#             u1 <= r1 + eps or u2 <= r2 + eps for u1, u2 in front_utils
#         )
#         p.run()
#         assert len(p.history) > 0, f"{p.state}"
#         if (neg, neg) in COMPLETE:
#             assert (
#                 p.agreement is not None or no_valid_outcomes
#             ), f"No agreement in a supposedly complete profile {_history(p)}{_plot(p, True)}"
#         if (neg, neg) in OPTIMAL:
#             assert (
#                 p.agreement in front_outcomes or p.agreement is None
#             ), f"Suboptimal agreement in a supposedly optimal profile {_history(p)}{_plot(p, True)}"
#
#
# def run_adversarial_case(
#     buyer,
#     seller,
#     normalized=False,
#     seller_reserved=0.0,
#     buyer_reserved=0.0,
#     cardinality=INFINITE,
#     mechanism_type: type[Mechanism] = TAUMechanism,
#     force_plot=False,
#     do_asserts=True,
#     n_steps=10 * 10 * 3,
# ):
#     n1, n2 = 100, 11
#     # create negotiation agenda (issues)
#     issues = [
#         make_issue(name="issue", values=n1 + n2),
#     ]
#     os = make_os(issues)
#
#     # create the mechanism
#     if mechanism_type == GeneralizedTAUMechanism:
#         session = mechanism_type(outcome_space=os, cardinality=cardinality)  # type: ignore
#     elif mechanism_type == TAUMechanism:
#         session = mechanism_type(outcome_space=os)
#     else:
#         session = mechanism_type(outcome_space=os, n_steps=n_steps)
#
#     # define buyer and seller utilities
#     seller_utility = MappingUtilityFunction(
#         dict(zip(os.enumerate(), [i for i in range(n1 + n2)])),  # type: ignore
#         outcome_space=os,
#     )
#     buyer_utility = MappingUtilityFunction(
#         dict(
#             zip(
#                 os.enumerate(),  # type: ignore
#                 [n1 + n2 for _ in range(n1)] + [n1 + n2 - 1 - i for i in range(n2)],
#             )
#         ),
#         outcome_space=os,
#     )
#     if normalized:
#         seller_utility = seller_utility.scale_max(1.0)
#         buyer_utility = buyer_utility.scale_max(1.0)
#     seller_utility.reserved_value = seller_reserved
#     buyer_utility.reserved_value = buyer_reserved
#
#     # create and add buyer and seller negotiators
#     b = buyer(name="buyer", id="buyer")
#     s = seller(name="seller", id="seller")
#     s.name += f"{s.short_type_name}"
#     b.name += f"{b.short_type_name}"
#     session.add(b, ufun=buyer_utility)
#     session.add(s, ufun=seller_utility)
#
#     front_utils, front_outcomes = session.pareto_frontier()
#     eps = 1e-3
#     no_valid_outcomes = all(
#         u1 <= buyer_reserved + eps or u2 <= seller_reserved + eps
#         for u1, u2 in front_utils
#     )
#     session.run()
#     assert len(session.history) > 0, f"{session.state}"
#     if do_asserts:
#         if (buyer, seller) in COMPLETE:
#             assert (
#                 session.agreement is not None or no_valid_outcomes
#             ), f"No agreement in a supposedly complete profile\n{_history(session)}{_plot(session, True, force=force_plot)}"
#         if (buyer, seller) in OPTIMAL:
#             assert (
#                 session.agreement in front_outcomes or session.agreement is None
#             ), f"Suboptimal agreement in a supposedly optimal profile\n{_history(session)}{_plot(session, True, force=force_plot)}"
#     _plot(session, force=force_plot)
#     return session
#
#
# def run_buyer_seller(
#     buyer,
#     seller,
#     normalized=False,
#     seller_reserved=0.0,
#     buyer_reserved=0.0,
#     cardinality=INFINITE,
#     min_unique=0,
#     mechanism_type: type[Mechanism] = TAUMechanism,
#     force_plot=False,
#     do_asserts=True,
#     n_steps=10 * 10 * 3,
#     mechanism_params: dict | None = None,
# ):
#     if not mechanism_params:
#         mechanism_params = dict()
#     # create negotiation agenda (issues)
#     issues = (
#         make_issue(name="price", values=10),
#         make_issue(name="quantity", values=(1, 11)),
#         make_issue(name="delivery_time", values=["today", "tomorrow", "nextweek"]),
#     )
#
#     # create the mechanism
#     if mechanism_type == GeneralizedTAUMechanism:
#         mechanism_params.update(dict(cardinality=cardinality, min_unique=min_unique))
#     elif mechanism_type == TAUMechanism:
#         pass
#     else:
#         mechanism_params.update(dict(n_steps=n_steps))
#     session = mechanism_type(issues=issues, **mechanism_params)  # type: ignore
#
#     # define buyer and seller utilities
#     seller_utility = LU(
#         values=[  # type: ignore
#             IdentityFun(),
#             LinearFun(0.2),
#             TableFun(dict(today=1.0, tomorrow=0.2, nextweek=0.0)),
#         ],
#         outcome_space=session.outcome_space,
#     )
#
#     buyer_utility = LU(
#         values={  # type: ignore
#             "price": AffineFun(-1, bias=9.0),
#             "quantity": LinearFun(0.2),
#             "delivery_time": TableFun(dict(today=0, tomorrow=0.7, nextweek=1.0)),
#         },
#         outcome_space=session.outcome_space,
#     )
#     if normalized:
#         seller_utility = seller_utility.scale_max(1.0)
#         buyer_utility = buyer_utility.scale_max(1.0)
#     seller_utility.reserved_value = seller_reserved
#     buyer_utility.reserved_value = buyer_reserved
#
#     # create and add buyer and seller negotiators
#     b = buyer(name="buyer", id="buyer")
#     s = seller(name="seller", id="seller")
#     s.name += f"{s.short_type_name}"
#     b.name += f"{b.short_type_name}"
#     session.add(b, ufun=buyer_utility)
#     session.add(s, ufun=seller_utility)
#
#     agreement = session.run().agreement
#
#     front_utils, front_outcomes = session.pareto_frontier(sort_by_welfare=True)
#     eps = 1e-3
#     no_valid_outcomes = all(
#         u1 <= buyer_reserved + eps or u2 <= seller_reserved + eps
#         for u1, u2 in front_utils
#     )
#     assert len(session.history) > 0, f"{session.state}"
#     if do_asserts:
#         if (buyer, seller) in COMPLETE:
#             assert (
#                 agreement is not None or no_valid_outcomes
#             ), f"No agreement in a supposedly complete profile\n{_history(session)}{_plot(session, True, force=force_plot)}"
#         if (buyer, seller) in OPTIMAL:
#             x = tuple(float(u(agreement)) for u in [buyer_utility, seller_utility])
#             dominating = [
#                 (front_utils[_], front_outcomes[_])
#                 for _ in dominating_points(x, front_utils)
#             ]
#             assert (
#                 len(dominating) == 0 or agreement is None
#             ), f"Suboptimal agreement in a supposedly optimal profile\nAgreement:{agreement} (u={x}) is dominated by {dominating}\n{_history(session)}{_plot(session, True, force=force_plot)}"
#             assert (
#                 agreement in front_outcomes or agreement is None
#             ), f"Suboptimal agreement in a supposedly optimal profile\n{_history(session)}{_plot(session, True, force=force_plot)}"
#     _plot(session, force=force_plot)
#     return session
#
#
# def remove_under_line(
#     os: DiscreteCartesianOutcomeSpace,
#     ufuns: Iterable[UtilityFunction],
#     limits=((0.0, 0.8), (0.0, None)),
# ):
#     mxs = [_.max() for _ in ufuns]
#
#     limits = list(limits)
#     for i, (x, y) in enumerate(limits):
#         if y is None:
#             limits[i] = (x, mxs[i])
#     x1, x2 = limits[0]
#     y1, y2 = limits[1]
#
#     outcomes = list(os.enumerate())
#     _, frontier = pareto_frontier(list(ufuns), outcomes)
#     frontier = set(frontier)
#
#     accepted: list[Outcome] = []
#     for outcome in outcomes:
#         if outcome in frontier:
#             accepted.append(outcome)
#             continue
#         xA, yA = (_(outcome) for _ in ufuns)
#         v1 = (x2 - x1, y2 - y1)  # Vector 1
#         v2 = (x2 - xA, y2 - yA)  # Vector 2
#         xp = v1[0] * v2[1] - v1[1] * v2[0]  # Cross product
#         if xp > 0:
#             # s = (y2-y1) / (x2 - x1)
#             # assert yA < s * xA + y1 - s * x1
#             continue
#         accepted.append(outcome)
#     return accepted
#
#
# def run_anac_example(
#     first_type,
#     second_type,
#     mechanism_type: type[Mechanism] = TAUMechanism,
#     force_plot=False,
#     do_asserts=True,
#     domain_name="cameradomain",
#     mechanism_params=dict(),
#     single_issue=False,
#     remove_under=False,
# ):
#     src = pkg_resources.resource_filename(
#         "negmas", resource_name=f"tests/data/{domain_name}"
#     )
#     base_folder = src
#     domain = Scenario.from_genius_folder(Path(base_folder))
#     assert domain is not None
#     # create the mechanism
#
#     a1 = first_type(
#         name="a1",
#         id="a1",
#     )
#     a2 = second_type(
#         name="a2",
#         id="a2",
#     )
#     a1.name += f"{a1.short_type_name}"
#     a2.name += f"{a2.short_type_name}"
#     domain.normalize()
#     if single_issue or remove_under:
#         domain = domain.to_single_issue(randomize=True)
#         assert len(domain.issues) == 1
#     if remove_under:
#         outcomes = remove_under_line(domain.outcome_space, domain.ufuns)  # type: ignore
#         issue = CategoricalIssue(
#             values=[_[0] for _ in outcomes], name=domain.outcome_space.issues[0].name
#         )
#         domain.outcome_space = make_os([issue])
#
#         # domain.ufuns = [ MappingUtilityFunction(dict(zip(outcomes, [u(_) for _ in outcomes])), outcome_space=domain.outcome_space) for u in domain.ufuns ]
#     domain.mechanism_type = mechanism_type
#     domain.mechanism_params = mechanism_params
#     neg = domain.make_session([a1, a2], n_steps=domain.outcome_space.cardinality)
#     if neg is None:
#         raise ValueError(f"Failed to lead domain from {base_folder}")
#
#     _, front_outcomes = neg.pareto_frontier()
#     neg.run()
#     assert len(neg.history) > 0, f"{neg.state}"
#     if do_asserts:
#         assert (
#             neg.agreement is not None
#         ), f"{_history(neg)}{_plot(neg, True, force=force_plot)}"
#         assert neg.agreement in front_outcomes or neg.agreement is None
#     _plot(neg, force=force_plot)
#     return neg
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2"], list(itertools.product(NEGOTIATORS, NEGOTIATORS))
# )
# def test_buyer_seller_easy(neg1, neg2):
#     run_buyer_seller(
#         neg1,
#         neg2,
#         normalized=True,
#         seller_reserved=0.1,
#         buyer_reserved=0.1,
#         force_plot=FORCE_PLOT,
#     )
#
#
# # @pytest.mark.skip(reason="Known failure. Enters an infinite loop")
# def test_buyer_seller_alphainf_war_cab():
#     run_buyer_seller(
#         WARNegotiator,
#         CABNegotiator,
#         mechanism_type=TAUMechanism,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         min_unique=0,
#         cardinality=INFINITE,
#     )
#
#
# # @pytest.mark.skip(reason="Known failure. Enters an infinite loop")
# def test_buyer_seller_easy_wab():
#     run_buyer_seller(
#         WABNegotiator,
#         WABNegotiator,
#         normalized=True,
#         seller_reserved=0.1,
#         buyer_reserved=0.1,
#         force_plot=FORCE_PLOT,
#     )
#
#
# # @pytest.mark.skip(reason="Known failure. Enters an infinite loop")
# def test_buyer_seller_easy_wan():
#     run_buyer_seller(
#         WANNegotiator,
#         WANNegotiator,
#         normalized=True,
#         seller_reserved=0.1,
#         buyer_reserved=0.1,
#         force_plot=FORCE_PLOT,
#     )
#
#
# def test_buyer_seller_easy_cab():
#     run_buyer_seller(
#         CABNegotiator,
#         CABNegotiator,
#         normalized=True,
#         seller_reserved=0.1,
#         buyer_reserved=0.1,
#         force_plot=FORCE_PLOT,
#     )
#
#
# def test_buyer_seller_gao_easy():
#     run_buyer_seller(
#         AspirationNegotiator,
#         AspirationNegotiator,
#         normalized=True,
#         seller_reserved=0.1,
#         buyer_reserved=0.1,
#         force_plot=FORCE_PLOT,
#         mechanism_type=GAOMechanism,
#         do_asserts=False,
#     )
#
#
# def test_buyer_seller_sao():
#     run_buyer_seller(
#         AspirationNegotiator,
#         AspirationNegotiator,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         mechanism_type=GAOMechanism,
#         do_asserts=False,
#     )
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2", "mechanism"],
#     list(
#         itertools.product(
#             NEGOTIATORS, NEGOTIATORS, (TAUMechanism, GeneralizedTAUMechanism)
#         )
#     ),
# )
# def test_buyer_seller_alphainf(neg1, neg2, mechanism):
#     run_buyer_seller(
#         neg1,
#         neg2,
#         mechanism_type=mechanism,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         min_unique=0,
#         cardinality=INFINITE,
#     )
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2"], list(itertools.product(NEGOTIATORS, NEGOTIATORS))
# )
# def test_buyer_seller_alpha0(neg1, neg2):
#     run_buyer_seller(
#         neg1,
#         neg2,
#         mechanism_type=GeneralizedTAUMechanism,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         min_unique=0,
#         cardinality=0,
#     )
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2"],
#     [
#         _
#         for _ in itertools.product(NEGOTIATORS, NEGOTIATORS)
#         if (_[0], _[1]) not in NORAISE
#     ],
# )
# def test_buyer_seller_betainf(neg1, neg2):
#     with pytest.raises(AssertionError):
#         run_buyer_seller(
#             neg1,
#             neg2,
#             mechanism_type=GeneralizedTAUMechanism,
#             normalized=True,
#             seller_reserved=0.5,
#             buyer_reserved=0.6,
#             force_plot=FORCE_PLOT,
#             min_unique=INFINITE,
#         )
#
#
# # @pytest.mark.parametrize(
# #     ["neg1", "neg2", "parallel", "accept_any"],
# #     [
# #         _
# #         for _ in itertools.product(
# #             NEGOTIATORS, NEGOTIATORS, [True, False], [True, False]
# #         )
# #         if (_[0], _[1]) not in NORAISE
# #     ],
# # )
# # def test_buyer_seller_tau_bilateral_any_acceptance_extension_passes(
# #     neg1, neg2, parallel, accept_any
# # ):
# #     run_buyer_seller(
# #         neg1,
# #         neg2,
# #         mechanism_type=TAUMechanism,
# #         normalized=True,
# #         seller_reserved=0.5,
# #         buyer_reserved=0.6,
# #         force_plot=FORCE_PLOT,
# #         min_unique=0,
# #         mechanism_params=dict(parallel=parallel, accept_in_any_thread=accept_any),
# #     )
#
#
# def test_buyer_seller_tau_bilateral_any_acceptance_extension_ok():
#     for neg1, neg2 in itertools.product(NEGOTIATORS, NEGOTIATORS):
#         if (neg1, neg2) in NORAISE:
#             continue
#         results = []
#         for parallel, accept_any in itertools.product([True, False], [True, False]):
#             x = run_buyer_seller(
#                 neg1,
#                 neg2,
#                 mechanism_type=TAUMechanism,
#                 normalized=True,
#                 seller_reserved=0.5,
#                 buyer_reserved=0.6,
#                 force_plot=FORCE_PLOT,
#                 min_unique=0,
#                 mechanism_params=dict(
#                     parallel=parallel, accept_in_any_thread=accept_any
#                 ),
#             )
#             if x is not None:
#                 results.append(x.agreement)
#             else:
#                 results.append(x)
#         assert all(a == b for a, b in zip(results[1:], results[:-1]))
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2", "mechanism"],
#     list(itertools.product(NEGOTIATORS, NEGOTIATORS, MECHS)),
# )
# def test_buyer_seller_beta0(neg1, neg2, mechanism):
#     run_buyer_seller(
#         neg1,
#         neg2,
#         mechanism_type=mechanism,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         min_unique=0,
#     )
#
#
# def test_buyer_seller_beta0_example_gtau():
#     run_buyer_seller(
#         CABNegotiator,
#         CABNegotiator,
#         mechanism_type=GeneralizedTAUMechanism,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         min_unique=0,
#     )
#
#
# def test_buyer_seller_beta0_example_tau():
#     run_buyer_seller(
#         CABNegotiator,
#         CABNegotiator,
#         mechanism_type=TAUMechanism,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         min_unique=0,
#     )
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2"], list(itertools.product(NEGOTIATORS, NEGOTIATORS))
# )
# def test_anac_scenario_example_single(neg1, neg2):
#     run_anac_example(
#         neg1,
#         neg2,
#         force_plot=FORCE_PLOT,
#         single_issue=True,
#         remove_under=False,
#     )
#
#
# def test_anac_scenario_example_gao_single():
#     run_anac_example(
#         AspirationNegotiator,
#         AspirationNegotiator,
#         mechanism_type=GAOMechanism,
#         force_plot=FORCE_PLOT,
#         single_issue=True,
#         remove_under=False,
#     )
#
#
# def test_anac_scenario_example_sao_single():
#     run_anac_example(
#         AspirationNegotiator,
#         AspirationNegotiator,
#         mechanism_type=SAOMechanism,
#         force_plot=FORCE_PLOT,
#         single_issue=True,
#         remove_under=False,
#     )
#
#
# @pytest.mark.skipif(not NEGMAS_RUN_GENIUS, reason="Skipping genius tests")
# def test_anac_scenario_example_genius_gao_single():
#     from negmas.genius.gnegotiators import Atlas3
#
#     run_anac_example(
#         Atlas3,
#         Atlas3,
#         mechanism_type=GAOMechanism,
#         force_plot=FORCE_PLOT,
#         single_issue=True,
#         remove_under=False,
#     )
#
#
# @pytest.mark.skipif(not NEGMAS_RUN_GENIUS, reason="Skipping genius tests")
# def test_anac_scenario_example_genius_sao_single():
#     from negmas.genius.gnegotiators import Atlas3
#
#     run_anac_example(
#         Atlas3,
#         Atlas3,
#         mechanism_type=SAOMechanism,
#         force_plot=FORCE_PLOT,
#         single_issue=True,
#         remove_under=False,
#     )
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2"], list(itertools.product(NEGOTIATORS, NEGOTIATORS))
# )
# def test_anac_scenario_example(neg1, neg2):
#     run_anac_example(neg1, neg2, force_plot=FORCE_PLOT)
#
#
# def test_anac_scenario_example_gao():
#     run_anac_example(
#         AspirationNegotiator,
#         AspirationNegotiator,
#         mechanism_type=GAOMechanism,
#         force_plot=FORCE_PLOT,
#     )
#
#
# @pytest.mark.skipif(not NEGMAS_RUN_GENIUS, reason="Skipping genius tests")
# def test_anac_scenario_example_genius():
#     from negmas.genius.gnegotiators import Atlas3
#
#     run_anac_example(
#         Atlas3,
#         Atlas3,
#         mechanism_type=SAOMechanism,
#         force_plot=FORCE_PLOT,
#     )
#
#
# @pytest.mark.parametrize(
#     ["neg1", "neg2"], list(itertools.product(NEGOTIATORS, NEGOTIATORS))
# )
# def test_adversarial_case_easy(neg1, neg2):
#     run_adversarial_case(
#         neg1,
#         neg2,
#         force_plot=FORCE_PLOT,
#     )
#
#
# @pytest.mark.skipif(not NEGMAS_RUN_GENIUS, reason="Skipping genius tests")
# def test_adversarial_case_gao_easy_genius():
#     from negmas.genius.gnegotiators import Atlas3
#
#     run_adversarial_case(
#         Atlas3,
#         Atlas3,
#         force_plot=FORCE_PLOT,
#         mechanism_type=GAOMechanism,
#         do_asserts=False,
#     )
#
#
# def test_adversarial_case_gao_easy():
#     run_adversarial_case(
#         AspirationNegotiator,
#         AspirationNegotiator,
#         force_plot=FORCE_PLOT,
#         mechanism_type=GAOMechanism,
#         do_asserts=False,
#     )
#
#
# @given(
#     neg=st.sampled_from(NEGOTIATORS),
#     r1=st.floats(0, 1),
#     r2=st.floats(0, 1),
#     n1=st.integers(3, 5),
#     n2=st.integers(3, 10),
#     U1=st.sampled_from(
#         [
#             LU,
#         ]
#     ),
#     U2=st.sampled_from(
#         [
#             LU,
#         ]
#     ),
# )
# @settings(deadline=10_000, verbosity=Verbosity.verbose, max_examples=10)
# def test_a_tau_session(neg, r1, r2, n1, n2, U1, U2):
#     eps = 1e-3
#     time.perf_counter()
#     os: DiscreteCartesianOutcomeSpace = make_os([make_issue(n1), make_issue(n2)])  # type: ignore
#     p = TAUMechanism(outcome_space=os)
#     ufuns = [U1.random(os, reserved_value=r1), U2.random(os, reserved_value=r2)]
#     for i, u in enumerate(ufuns):
#         p.add(
#             neg(name=f"CAB{i}"),
#             preferences=u,
#         )
#     p.run()
#     front_utils, front_outcomes = p.pareto_frontier()
#     no_valid_outcomes = all(u1 <= r1 + eps or u2 <= r2 + eps for u1, u2 in front_utils)
#     assert len(p.history) > 0, f"{p.state}"
#     if (neg, neg) in COMPLETE:
#         assert (
#             p.agreement is not None or no_valid_outcomes
#         ), f"No agreement in a supposedly complete profile\n{_history(p)}{_plot(p, True, force=False)}"
#     if (neg, neg) in OPTIMAL:
#         assert (
#             p.agreement in front_outcomes or p.agreement is None
#         ), f"Suboptimal agreement in a supposedly optimal profile\n{_history(p)}{_plot(p, True, force=False)}"
#
#
# from negmas.gb.adapters.tau import TAUNegotiatorAdapter, UtilityAdapter
#
#
# @given(
#     neg=st.sampled_from(SAONEGOTIATORS),
#     U1=st.sampled_from(
#         [
#             LU,
#         ]
#     ),
#     U2=st.sampled_from(
#         [
#             LU,
#         ]
#     ),
# )
# @settings(deadline=10_000, verbosity=Verbosity.verbose, max_examples=10)
# def test_a_tau_session_with_adapter(neg, U1, U2):
#     n1, n2 = 5, 10
#     time.perf_counter()
#     os: DiscreteCartesianOutcomeSpace = make_os([make_issue(n1), make_issue(n2)])  # type: ignore
#     p = TAUMechanism(outcome_space=os)
#     ufuns = [U1.random(os, reserved_value=0.0), U2.random(os, reserved_value=0.0)]
#     negotiator = neg(name=f"{neg.__name__}{0}")
#     p.add(TAUNegotiatorAdapter(base=negotiator), preferences=ufuns[0])
#     for i, u in enumerate(ufuns[1:]):
#         p.add(CABNegotiator(name=f"CAB{i}"), preferences=u)
#     p.run()
#     assert len(p.full_trace) > 0, f"{p.state}{_plot(p)}"
#
#
# @pytest.mark.skipif(not NEGMAS_RUN_GENIUS, reason="Skipping genius tests")
# def test_buyer_seller_gao_easy_genius():
#     from negmas.genius.gnegotiators import Atlas3
#
#     run_buyer_seller(
#         Atlas3,
#         Atlas3,
#         normalized=True,
#         seller_reserved=0.1,
#         buyer_reserved=0.1,
#         force_plot=FORCE_PLOT,
#         mechanism_type=GAOMechanism,
#         do_asserts=False,
#     )
#
#
# @pytest.mark.skipif(not NEGMAS_RUN_GENIUS, reason="Skipping genius tests")
# def test_buyer_seller_sao_genius():
#     from negmas.genius.gnegotiators import Atlas3
#
#     run_buyer_seller(
#         Atlas3,
#         Atlas3,
#         normalized=True,
#         seller_reserved=0.5,
#         buyer_reserved=0.6,
#         force_plot=FORCE_PLOT,
#         mechanism_type=SAOMechanism,
#         do_asserts=False,
#     )
#
#
# @given(
#     neg=st.sampled_from(NEGOTIATORS),
#     r1=st.floats(0, 1),
#     r2=st.floats(0, 1),
#     n1=st.integers(3, 5),
#     n2=st.integers(3, 10),
#     U1=st.sampled_from(
#         [
#             LU,
#         ]
#     ),
#     U2=st.sampled_from(
#         [
#             LU,
#         ]
#     ),
# )
# @settings(deadline=10_000, verbosity=Verbosity.verbose, max_examples=10)
# def test_tau_matches_generalized(neg, r1, r2, n1, n2, U1, U2):
#     eps = 1e-3
#     time.perf_counter()
#     os: DiscreteCartesianOutcomeSpace = make_os([make_issue(n1), make_issue(n2)])  # type: ignore
#     results = []
#     ufuns = [U1.random(os, reserved_value=r1), U2.random(os, reserved_value=r2)]
#     for cls in (TAUMechanism, GeneralizedTAUMechanism):
#         p = cls(outcome_space=os)
#         for i, u in enumerate(ufuns):
#             p.add(
#                 neg(name=f"CAB{i}"),
#                 preferences=u,
#             )
#         p.run()
#         front_utils, front_outcomes = p.pareto_frontier()
#         no_valid_outcomes = all(
#             u1 <= r1 + eps or u2 <= r2 + eps for u1, u2 in front_utils
#         )
#         assert len(p.history) > 0, f"{p.state}"
#         if (neg, neg) in COMPLETE:
#             assert (
#                 p.agreement is not None or no_valid_outcomes
#             ), f"No agreement in a supposedly complete profile\n{_history(p)}{_plot(p, True, force=False)}"
#         if (neg, neg) in OPTIMAL:
#             assert (
#                 p.agreement in front_outcomes or p.agreement is None
#             ), f"Suboptimal agreement in a supposedly optimal profile\n{_history(p)}{_plot(p, True, force=False)}"
#         results.append(dict(steps=p.current_step, agreement=p.agreement, trace=p.trace))
#     assert (
#         results[0]["agreement"] == results[1]["agreement"]
#     ), f"{results[0]['trace']=}\n{results[1]['trace']=}"
#     assert (
#         results[0]["steps"] == results[1]["steps"]
#     ), f"{results[0]['trace']=}\n{results[1]['trace']=}"
#
#
# @given(vals=st.lists(st.floats(), min_size=2, max_size=100), r=st.floats())
# @example(
#     vals=[0.0, 0.0],
#     r=1.0,
# )
# def test_tau_adapter_adapts_correctly(vals, r):
#     r = (r - 0.5) * 2.5
#     # vals = [0.0, 0.0, 1.0, 2.0, 3.0, 6.999, 7.0, 7.0, 7.1, 7.22, 8.0, 10]
#     # r = random.choice(vals)
#     os = make_os([make_issue(len(vals))])
#     outcomes = list(os.enumerate_or_sample())
#     ufun = MappingUtilityFunction(
#         dict(zip(outcomes, vals)), outcome_space=os, reserved_value=r
#     )
#     utiladapter = UtilityAdapter(ufun)
#     for i in range(len(outcomes) - 1, -1, -1):
#         assert utiladapter(outcomes[i]) == outcomes[i]
#     rational_outcomes = [_ for _ in outcomes if ufun(_) >= r]
#     if len(rational_outcomes):
#         sample = random.choices(rational_outcomes, k=2 * len(vals))
#         outcomes = set(outcomes)
#         utiladapter = UtilityAdapter(ufun)
#         for outcome in sample:
#             shouldbesame = outcome not in utiladapter.offered
#             shouldrepeat = len(utiladapter.offered) == len(outcomes)
#             x = utiladapter(outcome)
#             if shouldrepeat:
#                 assert (
#                     x == utiladapter.last_offer
#                 ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#             if shouldbesame:
#                 assert (
#                     x == outcome
#                 ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#             if ufun(outcome) >= ufun.reserved_value:
#                 if ufun(x) < ufun.reserved_value:
#                     pass
#                 assert (
#                     ufun(x) >= ufun.reserved_value
#                 ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#             assert (
#                 x in outcomes
#             ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#             assert (
#                 x in utiladapter.offered
#             ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#
#     # testing crazy startegy with some irrational outcomes
#     sample = random.choices(list(outcomes), k=100)
#     outcomes = set(outcomes)
#     utiladapter = UtilityAdapter(ufun)
#     for outcome in sample:
#         print(outcome)
#         shouldbesame = outcome not in utiladapter.offered
#         shouldrepeat = len(utiladapter.offered) == len(outcomes)
#         x = utiladapter(outcome)
#         if shouldrepeat:
#             assert (
#                 x == utiladapter.last_offer
#             ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#         if shouldbesame:
#             assert (
#                 x == outcome
#             ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#         if ufun(outcome) >= ufun.reserved_value:
#             if ufun(x) < ufun.reserved_value:
#                 pass
#             assert (
#                 ufun(x) >= ufun.reserved_value or x == utiladapter.last_offer
#             ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#         assert (
#             x in outcomes
#         ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#         assert (
#             x in utiladapter.offered
#         ), f"{x=}, {outcome=}, {utiladapter.last_offer=}, {utiladapter.offered=}"
#
#
# def test_a_tau_session_example_cab():
#     for _ in range(100):
#         r1, r2, n1, n2 = 0.2, 0.3, 2, 2
#         eps = 1e-3
#         time.perf_counter()
#         os: DiscreteCartesianOutcomeSpace = make_os([make_issue(n1), make_issue(n2)])  # type: ignore
#         p = TAUMechanism(outcome_space=os)
#         ufuns = [LU.random(os, reserved_value=r1), LU.random(os, reserved_value=r2)]
#         for i, u in enumerate(ufuns):
#             p.add(
#                 CABNegotiator(name=f"RCS{i}", id=f"RCS{i}"),
#                 preferences=u,
#             )
#         front_utils, front_outcomes = p.pareto_frontier()
#         no_valid_outcomes = all(
#             u1 <= r1 + eps or u2 <= r2 + eps for u1, u2 in front_utils
#         )
#         p.run()
#         assert len(p.history) > 0, f"{p.state}"
#         assert (
#             p.agreement is not None or no_valid_outcomes
#         ), f"No agreement in a supposedly complete profile {_history(p)}{_plot(p, True)}"
#         assert (
#             p.agreement in front_outcomes or p.agreement is None
#         ), f"Suboptimal agreement in a supposedly optimal profile {_history(p)}{_plot(p, True)}"



from __future__ import annotations
import math
from itertools import chain
from typing import TYPE_CHECKING

from ..crisp_ufun import UtilityFunction
from ..mixins import StationaryMixin
from .mapping import MappingUtilityFunction

if TYPE_CHECKING:
    from negmas.outcomes import Outcome

    from ..base_ufun import BaseUtilityFunction

__all__ = ["RankOnlyUtilityFunction"]


class RankOnlyUtilityFunction(StationaryMixin, UtilityFunction):
    r"""
    A utility function that keeps trak of outcome order onlyself.

    Remarks:
        - This type of utility function can only be generated for discrete outcome spaces.
        - It can be constructed from any utility function.
        - Given an outcome space of $K$ outcomes, the outcome with maximum utility will have a value $\le K$ when evaluated and the worst outcome will have value $0$.
        - Outcomes with equal utility value will be ordered randomly if `randomize_eqial` is `True`, otherwise outcomes that are `eps` apart will have the same rank.
        - The reserved value will also be mapped to an integer giving its rank

    """

    def eval(self, outcome: Outcome) -> int | None:
        return self._mapping.get(outcome, None)

    def __init__(
        self,
        ufun: BaseUtilityFunction,
        randomize_equal: bool = False,
        eps=1e-7,
        name: str | None = None,
        id: str | None = None,
        type_name: str | None = None,
    ):
        if ufun.outcome_space is None:
            raise ValueError(
                "Cannot craerte a RankOnly utility function for the given ufun because the outcome space is not konwn"
            )
        if not ufun.outcome_space.is_finite():
            raise ValueError(
                f"Cannot create a RankOnly utility function for the given ufun because the outcome space is not discrete\n{ufun.outcome_space}"
            )
        super().__init__(
            outcome_space=ufun.outcome_space, name=name, id=id, type_name=type_name
        )
        if math.isinf(ufun.reserved_value):
            outcomes = ufun.outcome_space.enumerate_or_sample()
        else:
            outcomes = chain(ufun.outcome_space.enumerate_or_sample(), [None])
        ordered: list[tuple[float, Outcome | None]] = [
            (float(ufun(_)), _) for _ in outcomes
        ]
        ordered = sorted(ordered, key=lambda x: x[0], reverse=True)
        ranked: list[tuple[Outcome | None, int]]
        if randomize_equal:
            n = ufun.outcome_space.cardinality - int(math.isinf(ufun.reserved_value))
            ranked = [(w, int(n - i)) for i, (_, w) in enumerate(ordered)]
        else:
            ranked = []
            last_ufun, current = float("inf"), -1
            for u, o in ordered:
                if abs(last_ufun - u) > eps:
                    current += 1
                last_ufun = u
                ranked.append((o, current))
            for i, (o, r) in enumerate(ranked):
                ranked[i] = (o, current - r)

        self._mapping = dict(ranked)
        if math.isinf(ufun.reserved_value):
            self.reserved_value = ufun.reserved_value
        else:
            self.reserved_value = self.eval(None)  # type: ignore

    def to_mapping_ufun(self) -> MappingUtilityFunction:
        return MappingUtilityFunction(
            self._mapping,
            outcome_space=self.outcome_space,
            reserved_value=self.reserved_value,
        )  # type: ignore



from __future__ import annotations
import numbers
import random
import sys

from negmas.outcomes.base_issue import Issue
from negmas.outcomes.contiguous_issue import ContiguousIssue
from negmas.outcomes.continuous_issue import ContinuousIssue

__all__ = ["CountableInfiniteIssue", "ContinuousInfiniteIssue", "InfiniteIssue"]

INFINITE_INT = sys.maxsize


class InfiniteIssue:
    """
    Indicates that the issue is infinite (i.e. one or more of its limits is infinity)
    """


class CountableInfiniteIssue(ContiguousIssue, InfiniteIssue):
    """
    An issue that can have all integer values.

    Remarks:
        - Actually, inifinties are replace with +- INFINITE_INT which is a very large number
    """

    def __init__(
        self, values: tuple[int | float, int | float], *args, **kwargs
    ) -> None:
        v: tuple[int, int] = tuple(  # type: ignore
            int(_)
            if isinstance(_, numbers.Integral)
            else INFINITE_INT
            if _ > 0
            else -INFINITE_INT
            for _ in values
        )
        super().__init__(v, *args, **kwargs)
        self._value_type = int
        self._n_values = float("inf")
        self.min_value, self.max_value = values

    def is_continuous(self) -> bool:
        return False

    def is_integer(self) -> bool:
        return True

    @property
    def cardinality(self) -> float:
        return float("inf")

    def rand_invalid(self):
        if self._values[0] == -INFINITE_INT and self._values[1] == INFINITE_INT:
            return None
        if self._values[0] == -INFINITE_INT:
            return random.randint(self._values[1] + 1, INFINITE_INT)
        return random.randint(-INFINITE_INT, self._values[0] - 1)

    def value_at(self, index: int):
        raise ValueError("Cannot index an infinite issue")

    def contains(self, issue: Issue) -> bool:
        """Checks weather this issue contains the input issue (i.e. every value in the input issue is in this issue)"""
        return issubclass(issue.value_type, numbers.Integral)


class ContinuousInfiniteIssue(ContinuousIssue, InfiniteIssue):
    """
    An issue that can represent all real numbers
    """

    def contains(self, issue: Issue) -> bool:
        """Checks weather this issue contains the input issue (i.e. every value in the input issue is in this issue)"""
        return issubclass(issue.value_type, numbers.Real)



from __future__ import annotations
from typing import TYPE_CHECKING

from attrs import define


def __setattr__(self, name, value):
    pass


def __delattr__(self, name):
    pass


if TYPE_CHECKING:
    from ...common import (
        MechanismState,
        NegotiatorMechanismInterface,
        PreferencesChange,
    )
    from ...negotiators import Negotiator
    from ...preferences import BaseUtilityFunction, Preferences

__all__ = ["Component"]


@define
class Component:
    """
    A component that can be added to a `ModularNegotiator`
    """

    _negotiator: Negotiator

    @property
    def negotiator(self):
        return self._negotiator

    def set_negotiator(self, negotiator: Negotiator) -> None:
        """
        Sets the negotiator of which this component is a part.
        """
        self._negotiator = negotiator

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        """
        Called to inform the component that the ufun has changed and the kinds of change that happened.
        """

    def can_join(
        self,
        nmi: NegotiatorMechanismInterface,
        state: MechanismState,
        *,
        preferences: Preferences | None = None,
        ufun: BaseUtilityFunction | None = None,
        role: str = "negotiator",
    ) -> bool:
        """
        A call back called before joining a negotiation to confirm that we can join it.
        """
        return True

    def after_join(self, nmi: NegotiatorMechanismInterface) -> None:
        """
        A call back called after joining a negotiation to confirm wwe joined.
        """

    def on_negotiation_start(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation start
        """

    def on_round_start(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation round start
        """

    def on_round_end(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation round end
        """

    def on_leave(self, state: MechanismState) -> None:
        """
        A call back called after leaving a negotiation.
        """

    def on_negotiation_end(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation end
        """

    def on_mechanism_error(self, state: MechanismState) -> None:
        """
        A call back called whenever an error happens in the mechanism. The error and its explanation are accessible in
        `state`
        """



from __future__ import annotations
import math
from abc import abstractmethod
from typing import Literal, Protocol, runtime_checkable

__all__ = ["TimeCurve", "Aspiration", "PolyAspiration", "ExpAspiration"]


@runtime_checkable
class TimeCurve(Protocol):
    """
    Models a time-curve mapping relative timge (going from 0.0 to 1.0) to a utility range to use
    """

    @abstractmethod
    def utility_range(self, t: float) -> tuple[float, float]:
        pass


@runtime_checkable
class Aspiration(TimeCurve, Protocol):
    """
    A monotonically decreasing time-curve
    """

    @abstractmethod
    def utility_at(self, t: float) -> float:
        pass

    def utility_range(self, t: float) -> tuple[float, float]:
        return self.utility_at(t), 1.0


class ExpAspiration(Aspiration):
    """
    An exponential conceding curve

    Args:
        max_aspiration: The aspiration level to start from (usually 1.0)
        aspiration_type: The aspiration type. Can be a string ("boulware", "linear", "conceder") or a number giving the exponent of the aspiration curve.
    """

    def __init__(
        self,
        max_aspiration: float,
        aspiration_type: Literal["boulware"]
        | Literal["conceder"]
        | Literal["linear"]
        | float,
    ):
        self.max_aspiration = max_aspiration
        self.aspiration_type = aspiration_type
        self.exponent = 1.0
        if isinstance(aspiration_type, int):
            self.exponent = float(aspiration_type)
        elif isinstance(aspiration_type, float):
            self.exponent = aspiration_type
        elif aspiration_type == "boulware":
            self.exponent = 0.125
        elif aspiration_type == "linear":
            self.exponent = 0.725
        elif aspiration_type == "conceder":
            self.exponent = 4.0
        else:
            raise ValueError(f"Unknown aspiration type {aspiration_type}")
        self._denominator = math.exp(1) - 1

    def utility_at(self, t: float) -> float:
        """
        The aspiration level

        Args:
            t: relative time (a number between zero and one)

        Returns:
            aspiration level
        """
        if t is None:
            raise ValueError(
                "Aspiration negotiators cannot be used in negotiations with no time or #steps limit!!"
            )
        return (
            self.max_aspiration
            * (math.exp(math.pow(1 - t, self.exponent)) - 1)
            / self._denominator
        )


class PolyAspiration(Aspiration):
    """
    A polynomially conceding curve

    Args:
        max_aspiration: The aspiration level to start from (usually 1.0)
        aspiration_type: The aspiration type. Can be a string ("boulware", "linear", "conceder") or a number giving the exponent of the aspiration curve.
    """

    def __init__(
        self,
        max_aspiration: float,
        aspiration_type: Literal["boulware"]
        | Literal["conceder"]
        | Literal["linear"]
        | Literal["hardheaded"]
        | float,
    ):
        self.max_aspiration = max_aspiration
        self.aspiration_type = aspiration_type
        self.exponent = 1.0
        if isinstance(aspiration_type, int):
            self.exponent = float(aspiration_type)
        elif isinstance(aspiration_type, float):
            self.exponent = aspiration_type
        elif aspiration_type == "boulware":
            self.exponent = 4.0
        elif aspiration_type == "linear":
            self.exponent = 1.0
        elif aspiration_type == "conceder":
            self.exponent = 0.25
        elif aspiration_type == "hardheaded":
            self.exponent = float("inf")
        else:
            raise ValueError(f"Unknown aspiration type {aspiration_type}")

    def utility_at(self, t: float) -> float:
        """
        The aspiration level

        Args:
            t: relative time (a number between zero and one)

        Returns:
            aspiration level
        """
        if t is None:
            raise ValueError(
                "Aspiration negotiators cannot be used in negotiations with no time or #steps limit!!"
            )
        return self.max_aspiration * (1.0 - math.pow(t, self.exponent))



"""
Implements `NambedObject` for entities that have name/id combinations.


In principle, the user should manage the `name` and use it for printing/logging.
The system manages the `id` making sure it is unique for every entity.
The user **can** override the system assigned `id` but should always make sure that
no entities in the entire system has te same `id`
"""

from __future__ import annotations

import datetime
import uuid
from os import PathLike
from pathlib import Path
from typing import Any, Literal, overload

import dill

from ..helpers import get_full_type_name, shorten, unique_name
from ..helpers.inout import dump, load

__all__ = ["NamedObject"]


class NamedObject:
    """The base class of all named entities.

    All named entities need to call this class's __init__() somewhere during initialization.

    Args:
        name (str): The given name of the entity. Notice that the class will add this to a  base that depends
                    on the child's class name.
        id (str): A unique identifier in the whole system. In principle you should let the system create
                  this identifier by passing None. In special cases like in serialization you may want to
                  set the id directly
        type_name (str): A string to be returned by `type_name` and its short version is returned by `short_type_name`

    """

    def __init__(
        self,
        name: str | None = None,
        *,
        id: str | None = None,
        type_name: str | None = None,
    ) -> None:
        if name is not None:
            name = str(name)
        self.__uuid = (
            (f"{name}-" if name is not None else "") + str(uuid.uuid4())
            if not id
            else id
        )
        if name is None or len(name) == 0:
            name = unique_name("", add_time=False, rand_digits=16)
        self.__name = name
        self.__type_name = type_name
        super().__init__()

    @classmethod
    def create(cls, *args, **kwargs):
        """Creates an object and returns a proxy to it."""
        return cls(*args, **kwargs)

    @classmethod
    def spawn_object(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    #     @classmethod
    #     def spawn_thread(cls, *args, **kwargs):
    #         raise NotImplementedError("Thread objects are not yet supported")
    #
    #     @classmethod
    #     def spawn_process(cls, *args, **kwargs):
    #         raise NotImplementedError("Process objects are not yet supported")
    #
    #     @classmethod
    #     def spawn_remote_tcp(
    #         cls,
    #         remote_tcp_address: str = "localhost",
    #         remote_tcp_port: Optional[int] = None,
    #         *args,
    #         **kwargs,
    #     ):
    #         raise NotImplementedError("TCP objects are not yet supported")
    #
    #     @classmethod
    #     def spawn_remote_http(
    #         cls,
    #         remote_http_address: str = "localhost",
    #         remote_http_port: Optional[int] = None,
    #         *args,
    #         **kwargs,
    #     ):
    #         raise NotImplementedError("HTTP objects are not yet supported")

    @classmethod
    def spawn(
        cls,
        spawn_as="object",
        spawn_params: dict[str, Any] | None = None,
        *args,
        **kwargs,
    ):
        if spawn_as == "object":
            return cls.spawn_object(*args, **kwargs)
        # if spawn_as == "thread":
        #     return cls.spawn_thread(*args, **kwargs)
        # if spawn_as == "gevent" or spawn_as == "green-thread":
        #     return cls.spawn_gevent(*args, **kwargs)
        # if spawn_as == "eventlet":
        #     return cls.spawn_eventlet(*args, **kwargs)
        # if spawn_as == "process":
        #     return cls.spawn_process(*args, **kwargs)
        if spawn_params is None:
            spawn_params = dict()
        # if spawn_as == "tcp":
        #     return cls.spawn_remote_tcp(
        #         spawn_params.get("address", "localhost"),
        #         spawn_params.get("port", None),
        #         *args,
        #         **kwargs,
        #     )
        raise ValueError(f"cannot spawn as {spawn_as}")

    @property
    def name(self):
        """A convenient name of the entity (intended primarily for printing/logging/debugging)."""
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name

    @property
    def uuid(self):
        """The unique ID of this entity"""
        return self.__uuid

    @uuid.setter
    def uuid(self, uuid):
        self.__uuid = uuid

    @property
    def id(self):
        """The unique ID of this entity"""
        return self.__uuid

    @id.setter
    def id(self, id):
        self.__uuid = id

    def checkpoint(
        self,
        path: PathLike,
        file_name: str | None = None,
        info: dict[str, Any] | None = None,
        exist_ok: bool = False,
        single_checkpoint: bool = True,
        step_attribs: tuple[str, ...] = (
            "current_step",
            "_current_step",
            "_Entity__current_step",
            "_step",
        ),
    ) -> Path:
        """
        Saves a checkpoint of the current object at  the given path.

        Args:

            path: Full path to a directory to store the checkpoint
            file_name: Name of the file to dump into. If not given, a unique name is created
            info: Information to save with the checkpoint (must be json serializable)
            exist_ok: If true, override existing dump
            single_checkpoint: If true, keep a single checkpoint for the last step
            step_attribs: Attributes to represent the time-step of the object. Any of the given attributes will be
                          used in the file name generated if single_checkpoint is False. If single_checkpoint is True, the
                          filename will not contain time-step information

        Returns:
            full path to the file used to save the checkpoint

        """
        if file_name is None:
            base_name = (
                f"{self.__class__.__name__.split('.')[-1].lower()}.{unique_name('', add_time=False, rand_digits=8, sep='-')}"
                f".{self.id.replace('/', '_')}"
            )
        else:
            base_name = file_name
        path = Path(path)
        if path.exists() and path.is_file():
            raise ValueError(f"{str(path)} is a file. It must be a directory")
        path.mkdir(parents=True, exist_ok=True)
        current_step = None
        for attrib in step_attribs:
            try:
                a = getattr(self, attrib)
                if isinstance(a, int):
                    current_step = a
                    break
            except AttributeError:
                pass
        if not single_checkpoint and current_step is not None:
            base_name = f"{current_step:05}.{base_name}"
        full_file_name = path / base_name

        if info is None:
            info = {}
        info.update(
            {
                "type": get_full_type_name(self.__class__),
                "id": self.id,
                "name": self.name,
                "time": datetime.datetime.now().isoformat(),
                "step": current_step,
                "filename": str(full_file_name),
            }
        )

        if (not exist_ok) and full_file_name.exists():
            raise ValueError(
                f"{str(full_file_name)} already exists. Pass exist_ok=True if you want to override it"
            )

        with open(full_file_name, "wb") as f:
            dill.dump(self, f)

        info_file_name = path / (base_name + ".json")
        dump(info, info_file_name)
        return full_file_name

    @overload
    @classmethod
    def from_checkpoint(
        cls, file_name: Path | str, return_info: Literal[False] = False
    ) -> NamedObject:
        ...

    @overload
    @classmethod
    def from_checkpoint(
        cls, file_name: Path | str, return_info: Literal[True] = True
    ) -> tuple[NamedObject, dict[str, Any]]:
        ...

    @classmethod
    def from_checkpoint(
        cls, file_name: Path | str, return_info: bool = False
    ) -> NamedObject | tuple[NamedObject, dict[str, Any]]:
        """
        Creates an object from a saved checkpoint

        Args:
            file_name:
            return_info: If True, tbe information saved when the file was dumped are returned

        Returns:
            Either the object or the object and dump-info as a dict (if return_info was true)

        Remarks:

            - If info is returned, it is guaranteed to have the following members:
                - time: Dump time
                - type: Type of the dumped object
                - id: ID
                - name: name
        """
        file_name = Path(file_name).absolute()
        with open(file_name, "rb") as f:
            obj = dill.load(f)
        if return_info:
            return obj, cls.checkpoint_info(file_name)
        return obj

    @classmethod
    def checkpoint_info(cls, file_name: Path | str) -> dict[str, Any]:
        """
        Returns the information associated with a dump of the object saved in the given file

        Args:
            file_name: Name of the object

        Returns:

        """
        file_name = Path(file_name).absolute()
        return load(file_name.parent / (file_name.name + ".json"))

    @property
    def type_name(self) -> str:
        if self.__type_name:
            return self.__type_name
        return get_full_type_name(self.__class__)

    @property
    def short_type_name(self) -> str:
        return shorten(self.type_name)



from __future__ import annotations

__all__ = ["Entity"]


class Entity:
    """Defines an entity that is a part of the world but does not participate in the simulation"""

    def __init__(self, type_postfix: str = ""):
        self._initialized = False
        self.__type_postfix = type_postfix
        self.__current_step = 0

    def _shorten(self, long_name):
        name = (
            long_name.split(".")[-1]
            .lower()
            .replace("factory_manager", "")
            .replace("manager", "")
        )
        name = (
            name.replace("factory", "")
            .replace("agent", "")
            .replace("miner", "")
            .replace("consumer", "")
        )
        if long_name.startswith("jnegmas"):
            name = f"j-{name}"
        name = name.strip("_")
        return name

    @classmethod
    def _type_name(cls):
        return cls.__module__ + "." + cls.__name__

    @property
    def type_name(self):
        """Returns the name of the type of this entity"""
        return self.__class__._type_name() + self.__type_postfix

    @property
    def short_type_name(self):
        """Returns a short name of the type of this entity"""
        return self._shorten(self.type_name)

    @property
    def type_postfix(self):
        return self.__type_postfix

    def init(self):
        """Override this method to modify initialization logic"""

    def init_(self):
        """Called to initialize the agent **after** the world is initialized. the AWI is accessible at this point."""
        self._initialized = True
        self.__current_step = 0
        self.init()

    def step(self):
        """Override this method to modify stepping logic"""

    def step_(self):
        """Called at every time-step. This function is called directly by the world."""
        if not self._initialized:
            self.init_()
        self.step()
        self.__current_step += 1



from __future__ import annotations
import random

from hypothesis import example, given
from hypothesis import strategies as st
from pytest_check import pytest

from negmas.helpers.misc import distribute_integer_randomly
from negmas.outcomes.outcome_space import CartesianOutcomeSpace
from negmas.preferences.crisp.linear import LinearAdditiveUtilityFunction
from negmas.preferences.crisp.mapping import MappingUtilityFunction
from negmas.preferences.generators import (
    GENERATOR_MAP,
    generate_multi_issue_ufuns,
    generate_single_issue_ufuns,
    generate_utility_values,
    make_curve_pareto,
    make_endpoints,
    make_non_pareto,
    make_pareto,
    make_piecewise_linear_pareto,
    make_zero_sum_pareto,
    sample_between,
)


def dominates(x, y):
    return any(a > b for a, b in zip(x, y))


@given(
    n_pareto=st.integers(2, 40),
    n_segments_min=st.integers(1, 10),
    n_segments_range=st.integers(0, 10),
)
@example(n_pareto=2, n_segments_min=2, n_segments_range=1)
def test_make_piecewise_pareto2(n_pareto, n_segments_min, n_segments_range):
    make_piecewise_linear_pareto(
        n_pareto,
        n_segments=(n_segments_min, n_segments_min + n_segments_range)
        if n_segments_range
        else n_segments_min,
    )


@given(n=st.integers(0, 100), m=st.integers(1, 200), min_per_bin=st.integers(0, 200))
def test_distribute_integer_randomly(n, m, min_per_bin):
    lst = distribute_integer_randomly(n, m)
    assert len(lst) == m
    assert sum(lst) == n


@given(n=st.integers(0, 100), m=st.integers(1, 200))
def test_distribute_integer_randomly_on_none(n, m):
    lst = distribute_integer_randomly(n, m, min_per_bin=None)
    assert len(lst) == m
    assert sum(lst) == n
    if lst:
        assert max(lst) - min(lst) <= 1


@given(
    start=st.floats(0.0, 1.0),
    rng=st.floats(0.0, 1.0),
    n=st.integers(1, 100),
    endpoint=st.booleans(),
    main_range_min=st.floats(0.0, 0.5),
    main_range_range=st.floats(0.0, 0.5),
)
def test_sample_between(start, rng, n, endpoint, main_range_min, main_range_range):
    main_range_min = round(main_range_min, 6)
    main_range_range = round(main_range_range, 6)
    end = start + rng
    lst = sample_between(
        start, end, n, endpoint, (main_range_min, main_range_min + main_range_range)
    )
    assert len(lst) == n
    assert all(start <= _ <= end for _ in lst)


@given(n_segments=st.integers(0, 100))
def test_make_endpoints(n_segments):
    points = make_endpoints(n_segments)
    prev = (float("-inf"), float("inf"))
    for point in points:
        assert len(point) == 2
        assert point[0] > prev[0]
        assert point[1] < prev[1]
        prev = point


@given(n_segments=st.integers(0, 100), n_outcomes=st.integers(0, 100))
@example(n_segments=3, n_outcomes=2)
def test_make_pareto(n_segments, n_outcomes):
    points = make_pareto(make_endpoints(n_segments), n_outcomes)
    assert len(points) == n_outcomes
    for i, p1 in enumerate(points):
        for j, p2 in enumerate(points):
            if i == j:
                continue
            assert (
                any(a < b for a, b in zip(p1, p2))
                or (all(a == b for a, b in zip(p1, p2)))
                or n_segments == 0
            )


@given(
    n_segments=st.integers(1, 100),
    n_pareto=st.integers(1, 100),
    n_non=st.integers(0, 400),
)
@example(n_segments=1, n_pareto=2, n_non=1)
def test_make_non_pareto(n_segments, n_pareto, n_non):
    pareto_points = make_pareto(make_endpoints(n_segments), n_pareto)
    assert len(pareto_points) == n_pareto
    points = make_non_pareto(pareto_points, n_non)
    assert len(points) == n_non

    for non_pareto in points:
        assert any(dominates(x, non_pareto) for x in pareto_points)


@given(n_pareto=st.integers(1, 100))
def test_make_zero_sum_pareto(n_pareto):
    points = make_zero_sum_pareto(n_pareto)
    assert len(points) == n_pareto
    for i, p1 in enumerate(points):
        for j, p2 in enumerate(points):
            if i == j:
                continue
            assert any(a < b for a, b in zip(p1, p2)) or (
                all(a == b for a, b in zip(p1, p2))
            )


@given(n_segments=st.integers(0, 100), n_pareto=st.integers(1, 100))
def test_make_piecewise_pareto(n_pareto, n_segments):
    points = make_piecewise_linear_pareto(n_pareto, n_segments=n_segments)
    assert len(points) == n_pareto
    for i, p1 in enumerate(points):
        for j, p2 in enumerate(points):
            if i == j:
                continue
            assert any(a < b for a, b in zip(p1, p2)) or (
                all(a == b for a, b in zip(p1, p2))
            )


@given(shape=st.floats(1e-3, 5.0), n_pareto=st.integers(1, 100))
def test_make_curve_pareto(n_pareto, shape):
    pareto = make_curve_pareto(n_pareto, shape=shape)
    assert len(pareto) == n_pareto
    for i, p1 in enumerate(pareto):
        for j, p2 in enumerate(pareto):
            if i == j:
                continue
            assert any(a < b for a, b in zip(p1, p2)) or (
                all(a == b for a, b in zip(p1, p2))
            )


@given(
    n_pareto=st.integers(1, 100),
    n_non=st.integers(0, 400),
    generator=st.sampled_from(list(GENERATOR_MAP.keys())),
)
def test_generate_utility_values(n_pareto, n_non, generator):
    n_outcomes = n_pareto + n_non
    points = generate_utility_values(
        n_pareto=n_pareto,
        n_outcomes=n_outcomes,
        pareto_first=True,
        pareto_generator=generator,
    )
    pareto = points[:n_pareto]
    non_paretos = points[n_pareto:]

    for i, p1 in enumerate(pareto):
        for j, p2 in enumerate(pareto):
            if i == j:
                continue
            assert any(a < b for a, b in zip(p1, p2)) or (
                all(a == b for a, b in zip(p1, p2))
            ), f"{p1} and {p2} should be Pareto optimal but neither dominates the other"
    for y in non_paretos:
        assert any(
            dominates(x, y) for x in pareto
        ), f"{y} is non-pareto but not dominated by any pareto outcome"


@pytest.mark.parametrize(
    "ufun_type, val_type, numeric, linear",
    [
        (LinearAdditiveUtilityFunction, int, True, True),
        (LinearAdditiveUtilityFunction, str, False, True),
        (MappingUtilityFunction, int, True, False),
        (MappingUtilityFunction, str, False, False),
    ],
)
def test_generate_singleissue_utility_values_example(
    ufun_type, val_type, numeric, linear
):
    ufuns = generate_single_issue_ufuns(
        n_pareto=100,
        n_outcomes=30,
        n_ufuns=2,
        pareto_generator=random.choice(list(GENERATOR_MAP.keys())),
        linear=linear,
        numeric=numeric,
    )
    assert len(ufuns) > 0
    assert ufuns[0].outcome_space is not None
    assert isinstance(ufuns[0].outcome_space, CartesianOutcomeSpace)
    assert all([isinstance(_, ufun_type) for _ in ufuns])
    outcome = ufuns[0].outcome_space.random_outcome()
    assert all([isinstance(_, val_type) for _ in outcome])


@pytest.mark.parametrize(
    "ufun_type, val_type, numeric, linear",
    [
        (LinearAdditiveUtilityFunction, int, True, True),
        (LinearAdditiveUtilityFunction, str, False, True),
        (MappingUtilityFunction, int, True, False),
        (MappingUtilityFunction, str, False, False),
    ],
)
def test_generate_multiissue_utility_values_example2(
    ufun_type, val_type, numeric, linear
):
    ufuns = generate_multi_issue_ufuns(
        5,
        n_values=10,
        sizes=None,
        n_ufuns=2,
        pareto_generators=tuple(GENERATOR_MAP.keys()),
        linear=linear,
        numeric=numeric,
    )
    assert len(ufuns) > 0
    assert ufuns[0].outcome_space is not None
    assert isinstance(ufuns[0].outcome_space, CartesianOutcomeSpace)
    assert all([isinstance(_, ufun_type) for _ in ufuns])
    outcome = ufuns[0].outcome_space.random_outcome()
    assert all([isinstance(_, val_type) for _ in outcome])


@given(
    fractions=st.lists(st.floats(0, 1), min_size=2, max_size=2),
    generator=st.sampled_from(list(GENERATOR_MAP.keys())),
)
def test_generate_singleissue_utility_rational_fraction(fractions, generator):
    ufuns = generate_single_issue_ufuns(
        n_pareto=100,
        n_outcomes=30,
        n_ufuns=2,
        pareto_generator=generator,
        rational_fractions=fractions,
        reservation_selector=lambda a, _: a,
    )
    for ufun, f in zip(ufuns, fractions):
        assert ufun.outcome_space
        outcomes = list(ufun.outcome_space.enumerate_or_sample())
        n_outcomes = ufun.outcome_space.cardinality
        assert (
            abs(
                int(f * n_outcomes + 0.5)
                - len(
                    [
                        v
                        for outcome in outcomes
                        if (v := float(ufun(outcome))) >= ufun.reserved_value
                    ]
                )
            )
            < 2
        )


@pytest.mark.parametrize(
    "ufun_type, val_type, numeric, linear",
    [
        (LinearAdditiveUtilityFunction, int, True, True),
        (LinearAdditiveUtilityFunction, str, False, True),
        (MappingUtilityFunction, int, True, False),
        (MappingUtilityFunction, str, False, False),
    ],
)
def test_generate_multiissue_utility_values_example(
    ufun_type, val_type, numeric, linear
):
    ufuns = generate_multi_issue_ufuns(
        5,
        n_values=10,
        sizes=None,
        n_ufuns=2,
        pareto_generators=tuple(GENERATOR_MAP.keys()),
        linear=linear,
        numeric=numeric,
    )
    assert len(ufuns) > 0
    assert ufuns[0].outcome_space is not None
    assert isinstance(ufuns[0].outcome_space, CartesianOutcomeSpace)
    assert all([isinstance(_, ufun_type) for _ in ufuns])
    outcome = ufuns[0].outcome_space.random_outcome()
    assert all([isinstance(_, val_type) for _ in outcome])



# -*- coding: utf-8 -*-
"""A framework for conducting multi-strand multilateral asynchronous negotiations on multiple issues."""


__author__ = """Yasser Mohammad"""
__email__ = "yasserfarouk@gmail.com"
__version__ = "0.10.16"


from .config import *
from .types import *
from .common import *
from .inout import *
from .mechanisms import *
from .negotiators import *
from .outcomes import *
from .gb import *
from .sao import *
from .situated import *
from .st import *
from .preferences import *
from .genius import *


__all__ = (
    config.__all__
    + types.__all__
    + common.__all__
    + outcomes.__all__
    + preferences.__all__
    + negotiators.__all__
    + mechanisms.__all__
    + gb.__all__
    + sao.__all__
    + st.__all__
    + inout.__all__
    + genius.__all__
    + situated.__all__
    # + modeling.__all__
    # + helpers.prob.__all__
    # + [
    #     "exceptions",
    #     "warnings",
    #     "generics",
    #     "helpers",
    #     "events",
    #     "tournaments",
    #     "elicitation",
    #     "helpers",
    # ]
)
